/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */

lexer grammar VerilogLexer;

import preprocessor=VerilogPreprocessor;

@header {
  package ru.ispras.verilog.parser.grammar;

  import ru.ispras.verilog.parser.VerilogFrontend;
  import ru.ispras.verilog.parser.VerilogLogger;
  import ru.ispras.verilog.parser.VerilogStandard;
}

@members {
  /** IEEE 1364-1995 reserved keywords. */
  private static final VerilogStandard IEEE_1364_1995 = VerilogStandard.IEEE_1364_1995;
  /** IEEE 1364-2001 reserved keywords (without configuration keywords). */
  private static final VerilogStandard IEEE_1364_2001_NC = VerilogStandard.IEEE_1364_2001_NC;
  /** IEEE 1364-2001 reserved keywords. */
  private static final VerilogStandard IEEE_1364_2001 = VerilogStandard.IEEE_1364_2001;
  /** IEEE 1364-2005 reserved keywords. */
  private static final VerilogStandard IEEE_1364_2005 = VerilogStandard.IEEE_1364_2005;

  /** Reference to the parent Verilog front-end. */
  private VerilogFrontend frontend;

  /**
   * Constructs the Verilog lexer.
   *
   * @param chars    the char stream to be tokenized.
   * @param frontend the reference to the parent Verilog front-end.
   */
  public VerilogLexer(final CharStream chars, final VerilogFrontend frontend) {
    this(chars);

    this.frontend = frontend;
    this.preprocessor.setFrontend(frontend);
  }

  /**
   * Skips a token if the scope is hidden.
   */
  private void pp() {
    if(frontend.isHidden()) {
      skip();
    }
  }

  /**
   * Checks whether a keyword is allowed.
   */
  private boolean check(final VerilogStandard standard) {
    return frontend.isAllowed(standard);
  }

  @Override
  public void emitErrorMessage(final String msg) {
    if (!frontend.isHidden()) {
      VerilogLogger.error(msg);
    }
  }
}

//--------------------------------------------------------------------------------------------------
// Whitespaces and Comments
//--------------------------------------------------------------------------------------------------

WS         : (SPACE)+        { skip(); };
NL         : (('\r')? '\n')+ { skip(); };
SL_COMMENT : '//' REST       { skip(); };
ML_COMMENT : '/*' .* '*/'    { skip(); };

fragment
SPACE : (' ' | '\t') ;

fragment
LINE : (~('\n' | '\r'))* ;

fragment
REST : LINE (NL | EOF) ;

//--------------------------------------------------------------------------------------------------
// Brackets and Delimiters
//--------------------------------------------------------------------------------------------------

LBRACK      : '['               { pp(); };
RBRACK      : ']'               { pp(); };

LCURLY      : '{'               { pp(); };
RCURLY      : '}'               { pp(); };

LPAREN      : '('               { pp(); };
RPAREN      : ')'               { pp(); };

COMMA       : ','               { pp(); };
SEMI        : ';'               { pp(); };
DOT         : '.'               { pp(); };
COLON       : ':'               { pp(); };

//--------------------------------------------------------------------------------------------------
// Operators
//--------------------------------------------------------------------------------------------------

AT          : '@'               { pp(); };
ASSIGN      : '='               { pp(); };
MINUS       : '-'               { pp(); };
SHARP       : '#'               { pp(); };
QMARK       : '?'               { pp(); };
PLUS        : '+'               { pp(); };
LNOT        : '!'               { pp(); };
BNOT        : '~'               { pp(); };
BAND        : '&'               { pp(); };
BNAND       : '~&'              { pp(); };
BOR         : '|'               { pp(); };
BNOR        : '~|'              { pp(); };
BXOR        : '^'               { pp(); };
BXNOR       : '~^' | '^~'       { pp(); };
STAR        : '*'               { pp(); };
POWER       : '**'              { pp(); };
DIV         : '/'               { pp(); };
MOD         : '%'               { pp(); };
EQ          : '=='              { pp(); };
NOTEQ       : '!='              { pp(); };
NOTEQCASE   : '!=='             { pp(); };
EQCASE      : '==='             { pp(); };
LAND        : '&&'              { pp(); };
LOR         : '||'              { pp(); };
LESS        : '<'               { pp(); };
LESSEQ      : '<='              { pp(); };
GREATER     : '>'               { pp(); };
GREATEREQ   : '>='              { pp(); };
LSR         : '>>'              { pp(); };
LSL         : '<<'              { pp(); };
ASR         : '>>>'             { pp(); };
ASL         : '<<<'             { pp(); };
TRIGGER     : '->'              { pp(); };
PPATH       : '=>'              { pp(); };
FPATH       : '*>'              { pp(); };
AND3        : '&&&'             { pp(); };

//--------------------------------------------------------------------------------------------------
// Keywords (IEEE 1364-1995)
//--------------------------------------------------------------------------------------------------

KW_ALWAYS              : { check(IEEE_1364_1995) }?    'always'              { pp(); };
KW_AND                 : { check(IEEE_1364_1995) }?    'and'                 { pp(); };
KW_ASSIGN              : { check(IEEE_1364_1995) }?    'assign'              { pp(); };
KW_BEGIN               : { check(IEEE_1364_1995) }?    'begin'               { pp(); };
KW_BUF                 : { check(IEEE_1364_1995) }?    'buf'                 { pp(); };
KW_BUFIF0              : { check(IEEE_1364_1995) }?    'bufif0'              { pp(); };
KW_BUFIF1              : { check(IEEE_1364_1995) }?    'bufif1'              { pp(); };
KW_CASE                : { check(IEEE_1364_1995) }?    'case'                { pp(); };
KW_CASEX               : { check(IEEE_1364_1995) }?    'casex'               { pp(); };
KW_CASEZ               : { check(IEEE_1364_1995) }?    'casez'               { pp(); };
KW_CMOS                : { check(IEEE_1364_1995) }?    'cmos'                { pp(); };
KW_DEASSIGN            : { check(IEEE_1364_1995) }?    'deassign'            { pp(); };
KW_DEFAULT             : { check(IEEE_1364_1995) }?    'default'             { pp(); };
KW_DEFPARAM            : { check(IEEE_1364_1995) }?    'defparam'            { pp(); };
KW_DISABLE             : { check(IEEE_1364_1995) }?    'disable'             { pp(); };
KW_EDGE                : { check(IEEE_1364_1995) }?    'edge'                { pp(); };
KW_ELSE                : { check(IEEE_1364_1995) }?    'else'                { pp(); };
KW_END                 : { check(IEEE_1364_1995) }?    'end'                 { pp(); };
KW_ENDCASE             : { check(IEEE_1364_1995) }?    'endcase'             { pp(); };
KW_ENDFUNCTION         : { check(IEEE_1364_1995) }?    'endfunction'         { pp(); };
KW_ENDMODULE           : { check(IEEE_1364_1995) }?    'endmodule'           { pp(); };
KW_ENDPRIMITIVE        : { check(IEEE_1364_1995) }?    'endprimitive'        { pp(); };
KW_ENDSPECIFY          : { check(IEEE_1364_1995) }?    'endspecify'          { pp(); };
KW_ENDTABLE            : { check(IEEE_1364_1995) }?    'endtable'            { pp(); };
KW_ENDTASK             : { check(IEEE_1364_1995) }?    'endtask'             { pp(); };
KW_EVENT               : { check(IEEE_1364_1995) }?    'event'               { pp(); };
KW_FOR                 : { check(IEEE_1364_1995) }?    'for'                 { pp(); };
KW_FORCE               : { check(IEEE_1364_1995) }?    'force'               { pp(); };
KW_FOREVER             : { check(IEEE_1364_1995) }?    'forever'             { pp(); };
KW_FORK                : { check(IEEE_1364_1995) }?    'fork'                { pp(); };
KW_FUNCTION            : { check(IEEE_1364_1995) }?    'function'            { pp(); };
KW_HIGHZ0              : { check(IEEE_1364_1995) }?    'highz0'              { pp(); };
KW_HIGHZ1              : { check(IEEE_1364_1995) }?    'highz1'              { pp(); };
KW_IF                  : { check(IEEE_1364_1995) }?    'if'                  { pp(); };
KW_IFNONE              : { check(IEEE_1364_1995) }?    'ifnone'              { pp(); };
KW_INITIAL             : { check(IEEE_1364_1995) }?    'initial'             { pp(); };
KW_INOUT               : { check(IEEE_1364_1995) }?    'inout'               { pp(); };
KW_INPUT               : { check(IEEE_1364_1995) }?    'input'               { pp(); };
KW_INTEGER             : { check(IEEE_1364_1995) }?    'integer'             { pp(); };
KW_JOIN                : { check(IEEE_1364_1995) }?    'join'                { pp(); };
KW_LARGE               : { check(IEEE_1364_1995) }?    'large'               { pp(); };
KW_LARGE0              : { check(IEEE_1364_1995) }?    'large0'              { pp(); };
KW_LARGE1              : { check(IEEE_1364_1995) }?    'large1'              { pp(); };
KW_MACROMODULE         : { check(IEEE_1364_1995) }?    'macromodule'         { pp(); };
KW_MEDIUM              : { check(IEEE_1364_1995) }?    'medium'              { pp(); };
KW_MEDIUM0             : { check(IEEE_1364_1995) }?    'medium0'             { pp(); };
KW_MEDIUM1             : { check(IEEE_1364_1995) }?    'medium1'             { pp(); };
KW_MODULE              : { check(IEEE_1364_1995) }?    'module'              { pp(); };
KW_NAND                : { check(IEEE_1364_1995) }?    'nand'                { pp(); };
KW_NEGEDGE             : { check(IEEE_1364_1995) }?    'negedge'             { pp(); };
KW_NMOS                : { check(IEEE_1364_1995) }?    'nmos'                { pp(); };
KW_NOR                 : { check(IEEE_1364_1995) }?    'nor'                 { pp(); };
KW_NOT                 : { check(IEEE_1364_1995) }?    'not'                 { pp(); };
KW_NOTIF0              : { check(IEEE_1364_1995) }?    'notif0'              { pp(); };
KW_NOTIF1              : { check(IEEE_1364_1995) }?    'notif1'              { pp(); };
KW_OR                  : { check(IEEE_1364_1995) }?    'or'                  { pp(); };
KW_OUTPUT              : { check(IEEE_1364_1995) }?    'output'              { pp(); };
KW_PARAMETER           : { check(IEEE_1364_1995) }?    'parameter'           { pp(); };
KW_PMOS                : { check(IEEE_1364_1995) }?    'pmos'                { pp(); };
KW_POSEDGE             : { check(IEEE_1364_1995) }?    'posedge'             { pp(); };
KW_PRIMITIVE           : { check(IEEE_1364_1995) }?    'primitive'           { pp(); };
KW_PULL0               : { check(IEEE_1364_1995) }?    'pull0'               { pp(); };
KW_PULL1               : { check(IEEE_1364_1995) }?    'pull1'               { pp(); };
KW_PULLUP              : { check(IEEE_1364_1995) }?    'pullup'              { pp(); };
KW_PULLDOWN            : { check(IEEE_1364_1995) }?    'pulldown'            { pp(); };
KW_RCMOS               : { check(IEEE_1364_1995) }?    'rcmos'               { pp(); };
KW_REAL                : { check(IEEE_1364_1995) }?    'real'                { pp(); };
KW_REALTIME            : { check(IEEE_1364_1995) }?    'realtime'            { pp(); };
KW_REG                 : { check(IEEE_1364_1995) }?    'reg'                 { pp(); };
KW_RELEASE             : { check(IEEE_1364_1995) }?    'release'             { pp(); };
KW_REPEAT              : { check(IEEE_1364_1995) }?    'repeat'              { pp(); };
KW_RNMOS               : { check(IEEE_1364_1995) }?    'rnmos'               { pp(); };
KW_RPMOS               : { check(IEEE_1364_1995) }?    'rpmos'               { pp(); };
KW_RTRAN               : { check(IEEE_1364_1995) }?    'rtran'               { pp(); };
KW_RTRANIF0            : { check(IEEE_1364_1995) }?    'rtranif0'            { pp(); };
KW_RTRANIF1            : { check(IEEE_1364_1995) }?    'rtranif1'            { pp(); };
KW_SCALARED            : { check(IEEE_1364_1995) }?    'scalared'            { pp(); };
KW_SMALL               : { check(IEEE_1364_1995) }?    'small'               { pp(); };
KW_SMALL0              : { check(IEEE_1364_1995) }?    'small0'              { pp(); };
KW_SMALL1              : { check(IEEE_1364_1995) }?    'small1'              { pp(); };
KW_SPECIFY             : { check(IEEE_1364_1995) }?    'specify'             { pp(); };
KW_SPECPARAM           : { check(IEEE_1364_1995) }?    'specparam'           { pp(); };
KW_STRONG0             : { check(IEEE_1364_1995) }?    'strong0'             { pp(); };
KW_STRONG1             : { check(IEEE_1364_1995) }?    'strong1'             { pp(); };
KW_SUPPLY0             : { check(IEEE_1364_1995) }?    'supply0'             { pp(); };
KW_SUPPLY1             : { check(IEEE_1364_1995) }?    'supply1'             { pp(); };
KW_TABLE               : { check(IEEE_1364_1995) }?    'table'               { pp(); };
KW_TASK                : { check(IEEE_1364_1995) }?    'task'                { pp(); };
KW_TIME                : { check(IEEE_1364_1995) }?    'time'                { pp(); };
KW_TRAN                : { check(IEEE_1364_1995) }?    'tran'                { pp(); };
KW_TRANIF0             : { check(IEEE_1364_1995) }?    'tranif0'             { pp(); };
KW_TRANIF1             : { check(IEEE_1364_1995) }?    'tranif1'             { pp(); };
KW_TRI                 : { check(IEEE_1364_1995) }?    'tri'                 { pp(); };
KW_TRI0                : { check(IEEE_1364_1995) }?    'tri0'                { pp(); };
KW_TRI1                : { check(IEEE_1364_1995) }?    'tri1'                { pp(); };
KW_TRIAND              : { check(IEEE_1364_1995) }?    'triand'              { pp(); };
KW_TRIOR               : { check(IEEE_1364_1995) }?    'trior'               { pp(); };
KW_TRIREG              : { check(IEEE_1364_1995) }?    'trireg'              { pp(); };
KW_VECTORED            : { check(IEEE_1364_1995) }?    'vectored'            { pp(); };
KW_WAIT                : { check(IEEE_1364_1995) }?    'wait'                { pp(); };
KW_WAND                : { check(IEEE_1364_1995) }?    'wand'                { pp(); };
KW_WEAK0               : { check(IEEE_1364_1995) }?    'weak0'               { pp(); };
KW_WEAK1               : { check(IEEE_1364_1995) }?    'weak1'               { pp(); };
KW_WHILE               : { check(IEEE_1364_1995) }?    'while'               { pp(); };
KW_WIRE                : { check(IEEE_1364_1995) }?    'wire'                { pp(); };
KW_WOR                 : { check(IEEE_1364_1995) }?    'wor'                 { pp(); };
KW_XNOR                : { check(IEEE_1364_1995) }?    'xnor'                { pp(); };
KW_XOR                 : { check(IEEE_1364_1995) }?    'xor'                 { pp(); };

//--------------------------------------------------------------------------------------------------
// Additional Keywords (IEEE 1364-2001-noconfig)
//--------------------------------------------------------------------------------------------------

KW_AUTOMATIC           : { check(IEEE_1364_2001_NC) }? 'automatic'           { pp(); };
KW_ENDGENERATE         : { check(IEEE_1364_2001_NC) }? 'endgenerate'         { pp(); };
KW_GENERATE            : { check(IEEE_1364_2001_NC) }? 'generate'            { pp(); };
KW_GENVAR              : { check(IEEE_1364_2001_NC) }? 'genvar'              { pp(); };
KW_LOCALPARAM          : { check(IEEE_1364_2001_NC) }? 'localparam'          { pp(); };
KW_NOSHOWCANCELLED     : { check(IEEE_1364_2001_NC) }? 'noshowcancelled'     { pp(); };
KW_PULSESTYLE_ONDETECT : { check(IEEE_1364_2001_NC) }? 'pulsestyle_ondetect' { pp(); };
KW_PULSESTYLE_ONEVENT  : { check(IEEE_1364_2001_NC) }? 'pulsestyle_onevent'  { pp(); };
KW_SHOWCANCELLED       : { check(IEEE_1364_2001_NC) }? 'showcancelled'       { pp(); };
KW_SIGNED              : { check(IEEE_1364_2001_NC) }? 'signed'              { pp(); };
KW_UNSIGNED            : { check(IEEE_1364_2001_NC) }? 'unsigned'            { pp(); };

//--------------------------------------------------------------------------------------------------
// Additional Keywords (IEEE 1364-2001)
//--------------------------------------------------------------------------------------------------

KW_CELL                : { check(IEEE_1364_2001) }?    'cell'                { pp(); };
KW_CONFIG              : { check(IEEE_1364_2001) }?    'config'              { pp(); };
KW_DESIGN              : { check(IEEE_1364_2001) }?    'design'              { pp(); };
KW_ENDCONFIG           : { check(IEEE_1364_2001) }?    'endconfig'           { pp(); };
KW_INCDIR              : { check(IEEE_1364_2001) }?    'incdir'              { pp(); };
KW_INCLUDE             : { check(IEEE_1364_2001) }?    'include'             { pp(); };
KW_INSTANCE            : { check(IEEE_1364_2001) }?    'instance'            { pp(); };
KW_LIBLIST             : { check(IEEE_1364_2001) }?    'liblist'             { pp(); };
KW_LIBRARY             : { check(IEEE_1364_2001) }?    'library'             { pp(); };
KW_USE                 : { check(IEEE_1364_2001) }?    'use'                 { pp(); };

//--------------------------------------------------------------------------------------------------
// Additional Keywords (IEEE 1364-2005)
//--------------------------------------------------------------------------------------------------

KW_UWIRE               : { check(IEEE_1364_2005) }?    'uwire'               { pp(); };

//--------------------------------------------------------------------------------------------------
// Identifiers
//--------------------------------------------------------------------------------------------------

PATHPULSE_PARAM        : 'PATHPULSE$' (IDENT)?                                { pp(); };

SETUP_IDENT            : '$setup'                                             { pp(); };
HOLD_IDENT             : '$hold'                                              { pp(); };
SETUPHOLD_IDENT        : '$setuphold'                                         { pp(); };
RECOVERY_IDENT         : '$recovery'                                          { pp(); };
REMOVAL_IDENT          : '$removal'                                           { pp(); };
RECREM_IDENT           : '$recrem'                                            { pp(); };
SKEW_IDENT             : '$skew'                                              { pp(); };
TIMESKEW_IDENT         : '$timeskew'                                          { pp(); };
FULLSKEW_IDENT         : '$fullskew'                                          { pp(); };
PERIOD_IDENT           : '$period'                                            { pp(); };
WIDTH_IDENT            : '$width'                                             { pp(); };
NOCHANGE_IDENT         : '$nochange'                                          { pp(); };

IDENT         : ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'$'|'0'..'9')* { pp(); };
ESCAPED_IDENT : '\\' (~(' '|'\r'|'\t'|'\n'))* (' '|'\r'|'\t'|'\n')            { pp(); };
SYSTEM_IDENT  : '$' IDENT                                                     { pp(); };

//--------------------------------------------------------------------------------------------------
// Literals
//--------------------------------------------------------------------------------------------------

NUMBER : (((SIZE)? (SPACE)* BASE)=> BASE_NUMBER | NOBASE_NUMBER) { pp(); };

STRING : '"'
         (
              ('\\n' | '\\t' | '\\\\' | '\\"' | ('\\' OCTDIGIT OCTDIGIT OCTDIGIT))
           | ~('\\' | '"' | '\r' | '\n')
         )*
         '"' { pp(); };

//--------------------------------------------------------------------------------------------------

fragment
VOCAB : '\u0003'..'\u007f' ;

fragment
BASE_NUMBER : (SIZE)? (SPACE)* BASE (SPACE)* BASE_DIGIT (BASE_DIGIT | '_')* ;

fragment
SIZE : (DIGIT)+ ;

fragment
BASE : '\'' ('s' | 'S')? ('d' | 'D' | 'h' | 'H' | 'o' | 'O' | 'b' | 'B') ;

fragment
BASE_DIGIT : DIGIT | HEXDIGIT | 'x' | 'X' | 'z' | 'Z' | '?' ;

fragment
NOBASE_NUMBER : DIGIT (DIGIT | '_')* ( '.' (DIGIT | '_')* )? (EXPONENT)? ;

fragment
DIGIT : ('0'..'9') ;

fragment
OCTDIGIT : ('0'..'7') ;

fragment
HEXDIGIT : ('A'..'F' | 'a'..'f') ;

fragment
EXPONENT : ('e' | 'E') ('+' | '-')? ('0'..'9')+ ;

//--------------------------------------------------------------------------------------------------

INCORRECT_TOKEN : . { pp(); };

//--------------------------------------------------------------------------------------------------

