/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */

parser grammar VerilogParser;

options {
  k            = 3;
  tokenVocab   = VerilogLexer;
  output       = AST;
  ASTLabelType = CommonTree;
}

tokens {
  AST_ACTIVITY;
  AST_ACTIVITY_TYPE;
  AST_ASSIGN_STATEMENT;
  AST_ASSIGN_TYPE;
  AST_ASSIGNMENT;
  AST_ATTRIBUTE;
  AST_ATTRIBUTED_ITEM;
  AST_ATTRIBUTED_STATEMENT;
  AST_ATTRIBUTED_UNIT;
  AST_ATTRIBUTES;
  AST_AUTOMATIC;
  AST_BINARY_OPERATION;
  AST_BLOCK_GENERATE;
  AST_BLOCK_ITEMS;
  AST_BLOCK_STATEMENT;
  AST_BLOCK_TYPE;
  AST_CASE_GENERATE;
  AST_CASE_GENERATE_ITEM;
  AST_CASE_GENERATE_ITEMS;
  AST_CASE_STATEMENT;
  AST_CASE_STATEMENT_ITEM;
  AST_CASE_STATEMENT_ITEMS;
  AST_CASE_TYPE;
  AST_CONCATENATION;
  AST_CONTINUOUS_ASSIGNMENT;
  AST_DATA_TYPE;
  AST_DECLARATION;
  AST_DEFINE_PARAMETER;
  AST_DELAY;
  AST_DELAYED_STATEMENT;
  AST_DIMENSIONS;
  AST_DISABLE_STATEMENT;
  AST_DUMMY;
  AST_EDGE;
  AST_ELEMENT;
  AST_ELEMENT_TYPE;
  AST_EVENT;
  AST_EVENT_CONTROL;
  AST_EVENTS;
  AST_EXPRESSION;
  AST_EXPRESSIONS;
  AST_FUNCTION_CALL;
  AST_GENERATE;
  AST_GENERATE_ITEM;
  AST_GENERATE_ITEMS;
  AST_GENERATE_ITEMS;
  AST_IF_GENERATE;
  AST_IF_STATEMENT;
  AST_INSTANCE;
  AST_INSTANTIATION;
  AST_INTERFACE;
  AST_LOOP_GENERATE;
  AST_LOOP_STATEMENT;
  AST_LOOP_TYPE;
  AST_MINTYPMAX_EXPRESSION;
  AST_MODIFIERS;
  AST_MODULE;
  AST_MODULE_ITEMS;
  AST_MODULE_TYPE;
  AST_NAME;
  AST_NULL_STATEMENT;
  AST_NUMBER;
  AST_OPERATION;
  AST_PARAMETER_TYPE;
  AST_PARAMETRIZATION;
  AST_PATH;
  AST_PATH_DECLARATION;
  AST_PATH_DESCRIPTION;
  AST_PATH_IF;
  AST_PATH_ITEM;
  AST_PATH_REFERENCE;
  AST_PATH_THEN;
  AST_PATH_TYPE;
  AST_PORT;
  AST_PORT_CONNECTION;
  AST_PORT_CONNECTIONS;
  AST_PROCEDURE;
  AST_PROCEDURE_ITEMS;
  AST_PROCEDURE_TYPE;
  AST_PULSESTYLE;
  AST_PULSESTYLE_TYPE;
  AST_RANGE;
  AST_RANGE_TYPE;
  AST_REFERENCE;
  AST_REFERENCES;
  AST_REPEAT_EVENTS;
  AST_REPLICATION;
  AST_ROOT;
  AST_SELECTION;
  AST_SHOWCANCELLED;
  AST_SHOWCANCELLED_TYPE;
  AST_SPECIFY;
  AST_STATEMENT;
  AST_STATEMENTS;
  AST_STRENGTH;
  AST_STRING;
  AST_TABLE;
  AST_TABLE_EDGE;
  AST_TABLE_ENTRY;
  AST_TABLE_LEVELS;
  AST_TASK_STATEMENT;
  AST_TERNARY_OPERATION;
  AST_TRIGGER_STATEMENT;
  AST_TYPE;
  AST_UNARY_OPERATION;
  AST_WAIT_STATEMENT;
}

@header {
  package ru.ispras.verilog.parser.grammar;

  import ru.ispras.verilog.parser.VerilogLogger;
}

@members {
  @Override
  public void emitErrorMessage(final String msg) {
    VerilogLogger.error(msg);
  }
}

//==================================================================================================
// Grammar Start Rule
//==================================================================================================

startRule
    : source_code
;

//==================================================================================================
// A.1.2 Verilog Source Text
//==================================================================================================

source_code
    : (description)* EOF
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ROOT (description)*)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
description
    : attrs=attributes
      (
          item=module_declaration
        | item=udp_declaration
//      | item=config_declaration
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_UNIT
      $item
      ^(AST_ATTRIBUTES ($attrs)*)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
module_declaration
    : (type=KW_MODULE | type=KW_MACROMODULE) name=module_identifier
      (params=module_parameter_list)?
      (
          /* IEEE 1364-1995 format */
          (list_of_ports)=> iface=list_of_ports SEMI
          /* IEEE 1364-2001 format */
        | (ports=list_of_port_declarations)? SEMI
      )
      (items+=module_or_generate_item)*
      KW_ENDMODULE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_MODULE
        $name               // -> ^(AST_NAME ...)
        ^(AST_MODULE_TYPE   $type)
        ^(AST_INTERFACE    ($iface)?)
        ^(AST_MODULE_ITEMS ($params)? ($ports)? ($items)*)
    ) // AST_MODULE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.1.3 Module Parameters and Ports
//==================================================================================================

module_parameter_list
    : SHARP LPAREN parameter_declaration (COMMA parameter_declaration)* RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      parameter_declaration
      ^(AST_ATTRIBUTES /* null */)
    )+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_ports
    : LPAREN (port (COMMA port)*)? RPAREN
//--------------------------------------------------------------------------------------------------
 -> port*
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
port
    : ports=port_expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PORT
        ^(AST_NAME       /* null */)
        ^(AST_REFERENCES ($ports)?)
    ) // AST_PORT
//--------------------------------------------------------------------------------------------------
    | DOT name=port_identifier LPAREN (ports=port_expression)? RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PORT
        $name            // -> ^(AST_NAME ...)
        ^(AST_REFERENCES ($ports)?)
    ) // AST_PORT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
port_expression
    : port_reference
    | LCURLY port_reference (COMMA port_reference)* RCURLY
//--------------------------------------------------------------------------------------------------
 -> (port_reference)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
port_reference
    : name=port_identifier (LBRACK (dim=range_expression)? RBRACK)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_REFERENCE
        ^(AST_PATH
            ^(AST_PATH_ITEM
                $name // -> ^(AST_NAME ...)
                ^(AST_EXPRESSION /* null */)
            ) // AST_PATH_ITEM
        ) // AST_PATH
        ^(AST_SELECTION ($dim)?)
    ) // AST_REFERENCE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_port_declarations
    : LPAREN (port_declaration (COMMA port_declaration)*)? RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      port_declaration
      ^(AST_ATTRIBUTES /* null */)
    )*
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
module_or_generate_item
    : attrs=attributes
      (
          item=port_declaration       SEMI
        | item=parameter_declaration  SEMI
        | item=specparam_declaration  SEMI
        | item=localparam_declaration SEMI
        | item=event_declaration      SEMI
        | item=reg_declaration        SEMI
        | item=net_declaration        SEMI
        | item=genvar_declaration     SEMI
        | item=parameter_override     SEMI
        | item=continuous_assign      SEMI
        | item=module_instantiation   SEMI // gate_instantiation, udp_instantiation
        | item=function_declaration
        | item=task_declaration
        | item=initial_construct
        | item=always_construct
        | item=loop_generate
        | item=if_generate
        | item=case_generate
        | item=generate_region
        | item=specify_block
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      $item // There may be multiple items.
      ^(AST_ATTRIBUTES ($attrs)*)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
parameter_override
    : KW_DEFPARAM defparam_assignment (COMMA defparam_assignment)*
//--------------------------------------------------------------------------------------------------
 -> defparam_assignment+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.1.5 Configuration Source Text
//==================================================================================================

// TODO:

//==================================================================================================
// A.2.1.1 Module Parameter Declarations
//==================================================================================================

localparam_declaration
    : KW_LOCALPARAM (type=parameter_type | (signed=KW_SIGNED)? (dim=range)?)
      param_assignment (COMMA param_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        param_assignment     // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   KW_LOCALPARAM)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
parameter_declaration
    : KW_PARAMETER (type=parameter_type | (signed=KW_SIGNED)? (dim=range)?)
      param_assignment (COMMA param_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        param_assignment     // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   KW_PARAMETER)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
specparam_declaration
    : KW_SPECPARAM (dim=range)? specparam_assignment (COMMA specparam_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        specparam_assignment // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   KW_SPECPARAM)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  /* null */)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  /* null */)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
parameter_type
    : KW_INTEGER
    | KW_REAL
    | KW_REALTIME
    | KW_TIME
;

//==================================================================================================
// A.2.1.2 Port Declarations
//==================================================================================================

port_declaration
    : (dir=KW_INPUT | dir=KW_INOUT | dir=KW_OUTPUT)
        (type=port_type | reg=KW_REG)? (signed=KW_SIGNED)? (dim=range)?
          port_decl_assignment (COMMA port_decl_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        port_decl_assignment // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   $dir)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)? ($reg)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.1.3 Type Declarations
//==================================================================================================

//==================================================================================================
event_declaration
    : KW_EVENT event_array (COMMA event_array)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        event_array          // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   KW_EVENT)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  /* null */)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  /* null */)
            ^(AST_DIMENSIONS /* null */)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_array
    : name=event_identifier (dim=range)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS ($dim)*)
        ^(AST_EXPRESSION /* null */)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
reg_declaration
    : (
          type=KW_INTEGER | type=KW_REAL | type=KW_TIME | type=KW_REALTIME
        | reg=KW_REG (signed=KW_SIGNED)? (dim=range)?
      )
      variable_decl (COMMA variable_decl)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        variable_decl        // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   /* null */)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)? ($reg)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
net_declaration
    : net_type (strength)? (access=KW_VECTORED | access=KW_SCALARED)?
        (signed=KW_SIGNED)? (dim=range)?
          (delay)? net_decl_assignment (COMMA net_decl_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        net_decl_assignment  // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   /* null */)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  net_type)
            ^(AST_STRENGTH   (strength)?)
            ^(AST_MODIFIERS  ($access)? ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
        // TODO: (delay)?
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.2.1 Net and Variable Types
//==================================================================================================

port_type
    : net_type
    | KW_INTEGER
    | KW_TIME
    | KW_REAL
    | KW_REALTIME
;

//==================================================================================================
net_type
    : KW_SUPPLY0
    | KW_SUPPLY1
    | KW_TRI
    | KW_TRIAND
    | KW_TRIOR
    | KW_TRI0
    | KW_TRI1
    | KW_UWIRE
    | KW_WIRE
    | KW_WAND
    | KW_WOR
    | KW_TRIREG // NOTE: Not a net type (it is added to the net types to simplify the grammar).
;

//==================================================================================================
variable_decl
    : (
          name=variable_identifier (dim+=range)*
        | name=variable_identifier ASSIGN expr=expression
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS ($dim)*)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.2.2 Strengths
//==================================================================================================

strength
    : LPAREN strength_identifier (COMMA strength_identifier)? RPAREN
//--------------------------------------------------------------------------------------------------
 -> strength_identifier+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
strength_identifier
    : drive_strength_identifier
    | charge_strength_identifier
;

//==================================================================================================
drive_strength_identifier
    : KW_SUPPLY0
    | KW_SUPPLY1
    | KW_STRONG0
    | KW_STRONG1
    | KW_PULL0
    | KW_PULL1
    | KW_LARGE0
    | KW_LARGE1
    | KW_WEAK0
    | KW_WEAK1
    | KW_MEDIUM0
    | KW_MEDIUM1
    | KW_SMALL0
    | KW_SMALL1
    | KW_HIGHZ0
    | KW_HIGHZ1
;

//==================================================================================================
charge_strength_identifier
    : KW_SMALL
    | KW_MEDIUM
    | KW_LARGE
;

//==================================================================================================
// A.2.2.3 Delays
//==================================================================================================

delay
    : (
        SHARP delay_value
      | SHARP LPAREN (delay_value (COMMA delay_value)*) RPAREN
    )
//--------------------------------------------------------------------------------------------------
 -> (delay_value)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
delay_value
    : number
    | identifier
//--------------------------------------------------------------------------------------------------
 -> ^(AST_REFERENCE
        ^(AST_PATH
            ^(AST_PATH_ITEM
                identifier // -> ^(AST_NAME ...)
                ^(AST_EXPRESSION /* null */)
            ) // AST_PATH_ITEM
        ) // AST_PATH
        ^(AST_SELECTION /* null */)
    ) // AST_REFERENCE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.4 Declaration Assignments
//==================================================================================================

//==================================================================================================
port_decl_assignment
    : name=port_identifier (dim+=range)* (ASSIGN expr=expression)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS ($dim)*)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
defparam_assignment
    : name=hierarchical_parameter_identifier ASSIGN expr=mintypmax_expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DEFINE_PARAMETER
        $name // -> ^(AST_PATH ...)
        ^(AST_EXPRESSION $expr)
    ) // AST_DEFINE_PARAMETER
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
net_decl_assignment
    : name=net_identifier (dim+=range)* (ASSIGN expr=expression)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS ($dim)*)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
param_assignment
    : name=parameter_identifier ASSIGN expr=mintypmax_expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS /* null */)
        ^(AST_EXPRESSION $expr)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
specparam_assignment
    : name=specparam_identifier ASSIGN
        (
            (LPAREN)=> LPAREN expr=mintypmax_expression (COMMA mintypmax_expression)? RPAREN
          | expr=mintypmax_expression
        )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS /* null */)
        ^(AST_EXPRESSION $expr)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.5 Declaration Ranges
//==================================================================================================

range
    : LBRACK lhs=expression COLON rhs=expression RBRACK
//--------------------------------------------------------------------------------------------------
 -> ^(AST_RANGE
        ^(AST_RANGE_TYPE COLON)
        ^(AST_EXPRESSION $lhs)
        ^(AST_EXPRESSION $rhs)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.6 Function Declarations
//==================================================================================================

function_declaration
    : KW_FUNCTION (auto=KW_AUTOMATIC)?
          (type=function_return_type | (signed=KW_SIGNED)? dim=range)?
          name=function_identifier
        (ports=list_of_port_declarations)? SEMI
        ((function_item_declaration)=> decls+=function_item_declaration)*
          body=statement
      KW_ENDFUNCTION
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PROCEDURE
        $name                // -> ^(AST_NAME ...)
        ^(AST_PROCEDURE_TYPE KW_FUNCTION)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
        ^(AST_AUTOMATIC      ($auto)?)
        ^(AST_PROCEDURE_ITEMS ($ports)? ($decls)*)
        ^(AST_STATEMENT       $body)
    ) // AST_PROCEDURE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
function_item_declaration
    : (block_item_declaration)=> block_item_declaration
//--------------------------------------------------------------------------------------------------
// -> ^(AST_ATTRIBUTED_ITEM ...)
//--------------------------------------------------------------------------------------------------
    | item=port_declaration SEMI
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      $item // -> ^(AST_DECLARATION ...)+
      ^(AST_ATTRIBUTES /* null */)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
function_return_type
    : KW_INTEGER
    | KW_REAL
    | KW_REALTIME
    | KW_TIME
;

//==================================================================================================
// A.2.7 Task Declarations
//==================================================================================================

task_declaration
    : KW_TASK (auto=KW_AUTOMATIC)? name=task_identifier
        (ports=list_of_port_declarations)? SEMI
        ((task_item_declaration)=> decls+=task_item_declaration)*
          body=statement
      KW_ENDTASK
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PROCEDURE
        $name                 // -> ^(AST_NAME ...)
        ^(AST_PROCEDURE_TYPE KW_TASK)
        ^(AST_TYPE
            ^(AST_DATA_TYPE   /* null */)
            ^(AST_STRENGTH    /* null */)
            ^(AST_MODIFIERS   /* null */)
            ^(AST_DIMENSIONS  /* null */)
        ) // AST_TYPE
        ^(AST_AUTOMATIC       ($auto)?)
        ^(AST_PROCEDURE_ITEMS ($ports)? ($decls)*)
        ^(AST_STATEMENT       $body)
    ) // AST_PROCEDURE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
task_item_declaration
    : (block_item_declaration)=> block_item_declaration
//--------------------------------------------------------------------------------------------------
// -> ^(AST_ATTRIBUTED_ITEM ...)
//--------------------------------------------------------------------------------------------------
    | item=port_declaration SEMI
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      $item // -> ^(AST_DECLARATION ...)+
      ^(AST_ATTRIBUTES /* null */)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.2.8 Block Item Declarations
//==================================================================================================

block_item_declaration
    : attrs=attributes
      (
          item=block_variable_declaration SEMI
        | item=localparam_declaration     SEMI
        | item=parameter_declaration      SEMI
        | item=event_declaration          SEMI
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      $item // -> ^(AST_DECLARATION ...)+
      ^(AST_ATTRIBUTES ($attrs)*)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
block_variable_declaration
    : (type=block_variable_type | reg=KW_REG (signed=KW_SIGNED)? (dim=range)?)
      block_variable_array (COMMA block_variable_array)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        block_variable_array // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE   /* null */)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  ($type)? ($reg)?)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  ($signed)?)
            ^(AST_DIMENSIONS ($dim)?)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
block_variable_array
    : name=variable_identifier (dim=range)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS ($dim)*)
        ^(AST_EXPRESSION /* null */)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

block_variable_type
    : KW_INTEGER
    | KW_REAL
    | KW_REALTIME
    | KW_TIME
;

//==================================================================================================
// A.4.1 Module Instantiation
//==================================================================================================

module_instantiation // gate_instantiation, udp_instantiation
    : (name=module_identifier | name=gate_identifier)
      (
          (delay)=> delay
        | (strength)=> strength (delay)?
        | SHARP LPAREN params=list_of_parameter_assignments RPAREN
      )?
      module_instance (COMMA module_instance)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_INSTANTIATION
        $name                 // -> ^(AST_NAME ...)
        ^(AST_STRENGTH        (strength)?)
        ^(AST_DELAY           (delay)?)
        ^(AST_PARAMETRIZATION ($params)?)
        module_instance       // -> ^(AST_INSTANCE ...)
    )+ // AST_INSTANTIATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_parameter_assignments
    : parameter_assignment (COMMA parameter_assignment)*
//--------------------------------------------------------------------------------------------------
 -> (parameter_assignment)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
parameter_assignment
    : expr=mintypmax_expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES /* null */)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
    | DOT name=parameter_identifier LPAREN (expr=mintypmax_expression)? RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES
            ^(AST_REFERENCE
                ^(AST_PATH
                    ^(AST_PATH_ITEM
                        $name // -> ^(AST_NAME ...)
                        ^(AST_EXPRESSION /* null */)
                    ) // AST_PATH_ITEM
                ) // AST_PATH
                ^(AST_SELECTION /* null */)
            ) // AST_REFERENCE
        ) // AST_REFERENCES
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
module_instance
@init { boolean named = false; }
    : ({named=true;} name=module_instance_identifier (dim=range)?)?
        connections=list_of_port_connections
//--------------------------------------------------------------------------------------------------
 -> {named}?
    ^(AST_INSTANCE
        $name                  // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS       ($dim)?)
        ^(AST_PORT_CONNECTIONS $connections)
    ) // AST_INSTANCE
 ->
    ^(AST_INSTANCE
        ^(AST_NAME             /* null */)
        ^(AST_DIMENSIONS       ($dim)?)
        ^(AST_PORT_CONNECTIONS $connections)
    ) // AST_INSTANCE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_port_connections
    : LPAREN (port_connection (COMMA port_connection)*) RPAREN
//--------------------------------------------------------------------------------------------------
// NOTE: Empty list of port connections is enabled, but it is represented as (<empty>).
 -> (port_connection)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
port_connection
@init { boolean named = false; }
    : attrs=attributes
      (
          {named=false;} (expr=expression)?
        | {named=true;} DOT name=port_identifier (LPAREN (expr=expression)? RPAREN)?
      )
//--------------------------------------------------------------------------------------------------
 -> {named}?
    ^(AST_PORT_CONNECTION
        $name            // -> ^(AST_NAME ...)
        ^(AST_EXPRESSION ($expr)?)
        ^(AST_ATTRIBUTES ($attrs)*)
    ) // AST_PORT_CONNECTION
 ->
    ^(AST_PORT_CONNECTION
        ^(AST_NAME       /* null */)
        ^(AST_EXPRESSION ($expr)?)
        ^(AST_ATTRIBUTES ($attrs)*)
    ) // AST_PORT_CONNECTION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.4.2 Generate Construct
//==================================================================================================

generate_region
    : KW_GENERATE
        (module_or_generate_item)*
      KW_ENDGENERATE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_GENERATE (module_or_generate_item)*)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
genvar_declaration
    : KW_GENVAR genvar_decl (COMMA genvar_decl)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DECLARATION
        genvar_decl         // -> ^(AST_ELEMENT ...)
        ^(AST_ELEMENT_TYPE  KW_GENVAR)
        ^(AST_TYPE
            ^(AST_DATA_TYPE  /* null */)
            ^(AST_STRENGTH   /* null */)
            ^(AST_MODIFIERS  /* null */)
            ^(AST_DIMENSIONS /* null */)
        ) // AST_TYPE
    )+ // AST_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
genvar_decl
    : name=genvar_identifier
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ELEMENT
        $name // -> ^(AST_NAME ...)
        ^(AST_DIMENSIONS /* null */)
        ^(AST_EXPRESSION /* null */)
    ) // AST_ELEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
loop_generate
    : KW_FOR LPAREN init=genvar_assignment SEMI cond=expression SEMI step=genvar_assignment RPAREN
      body=generate_block
//--------------------------------------------------------------------------------------------------
 -> ^(AST_LOOP_GENERATE
        $init               // -> ^(AST_ASSIGNMENT ...)
        ^(AST_EXPRESSION    $cond)
        $step               // -> ^(AST_ASSIGNMENT ...)
        ^(AST_GENERATE_ITEM $body)
    ) // AST_LOOP_GENERATE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
genvar_assignment
    : name=genvar_identifier ASSIGN expr=expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES
            ^(AST_REFERENCE
                ^(AST_PATH
                    ^(AST_PATH_ITEM
                        $name // -> ^(AST_NAME ...)
                        ^(AST_EXPRESSION /* null */)
                    ) // AST_PATH_ITEM
                ) // AST_PATH
                ^(AST_SELECTION /* null */)
            ) // AST_REFERENCE
        ) // AST_REFERENCES
        ^(AST_EXPRESSION $expr)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
if_generate
    : KW_IF LPAREN cond=expression RPAREN
        then_body=generate_block_or_null
        // NOTE: IEEE 1364-2005 Std grammar has been extended to support 'else if'.
        ((KW_ELSE)=> KW_ELSE else_body=generate_block_or_null)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_IF_GENERATE
        ^(AST_EXPRESSION     $cond)
        ^(AST_GENERATE_ITEM  $then_body)
        ^(AST_GENERATE_ITEM ($else_body)?)
    ) // AST_IF_GENERATE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
case_generate
    : KW_CASE LPAREN expr=expression RPAREN
        case_generate_item (case_generate_item)*
      KW_ENDCASE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CASE_GENERATE
        ^(AST_EXPRESSION $expr)
        ^(AST_CASE_GENERATE_ITEMS (case_generate_item)+)
    ) // AST_CASE_GENERATE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
case_generate_item
    : (
          match=list_of_expressions COLON body=generate_block_or_null
        | KW_DEFAULT (COLON)? body=generate_block_or_null
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CASE_GENERATE_ITEM
        ^(AST_EXPRESSIONS   ($match)?)
        ^(AST_GENERATE_ITEM ($body)?)
    ) // AST_CASE_GENERATE_ITEM
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
generate_block
@init { boolean named = false; }
    : module_or_generate_item
//--------------------------------------------------------------------------------------------------
 ->
    ^(AST_BLOCK_GENERATE
        ^(AST_NAME           /* null */)
        ^(AST_GENERATE_ITEMS module_or_generate_item)
    ) // AST_BLOCK_GENERATE
//--------------------------------------------------------------------------------------------------
    | KW_BEGIN
        (COLON {named=true;} name=generate_block_identifier)? (module_or_generate_item)*
      KW_END
//--------------------------------------------------------------------------------------------------
 -> {named}?
    ^(AST_BLOCK_GENERATE
        $name                // -> ^(AST_NAME ...)
        ^(AST_GENERATE_ITEMS (module_or_generate_item)*)
    ) // AST_BLOCK_GENERATE
 ->
    ^(AST_BLOCK_GENERATE
        ^(AST_NAME           /* null */)
        ^(AST_GENERATE_ITEMS (module_or_generate_item)*)
    ) // AST_BLOCK_GENERATE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
generate_block_or_null
    : generate_block
    | SEMI!
;

//==================================================================================================
// A.5.* UDP Declaration and Instantiation
//==================================================================================================

udp_declaration
    : (type=KW_PRIMITIVE) name=udp_identifier
      (
          /* IEEE 1364-1995 format */
          (list_of_ports)=> iface=list_of_ports SEMI
          /* IEEE 1364-2001 format */
        | (ports=list_of_port_declarations)? SEMI
      )
      (items+=udp_item)*
      KW_ENDPRIMITIVE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_MODULE
        $name              // -> ^(AST_NAME ...)
        ^(AST_MODULE_TYPE   $type)
        ^(AST_INTERFACE    ($iface)?)
        ^(AST_MODULE_ITEMS ($ports)? ($items)*)
    ) // AST_MODULE
//--------------------------------------------------------------------------------------------------
;

udp_item
    : attrs=attributes
      (
          item=port_declaration SEMI
        | item=reg_declaration  SEMI
        | item=initial_construct
        | item=udp_table
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_ITEM
      $item // There may be multiple items.
      ^(AST_ATTRIBUTES ($attrs)*)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
udp_table
    : KW_TABLE
        (table_entry)+
      KW_ENDTABLE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TABLE (table_entry)+)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
table_entry
    : (options {greedy=true;}: inputs1+=level_symbol)*
        (edge=edge_indicator
          (options {greedy=true;}: inputs2+=level_symbol)*)?
       COLON state1=level_symbol
      (COLON state2=level_symbol)? SEMI
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TABLE_ENTRY
        ^(AST_TABLE_LEVELS ($inputs1)*)
        ^(AST_TABLE_EDGE   ($edge)?)
        ^(AST_TABLE_LEVELS ($inputs2)*)
        ^(AST_TABLE_LEVELS ($state1)? ($state2)?)
    ) // AST_TABLE_ENTRY
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
edge_indicator
    : LPAREN! pair_of_level_symbols RPAREN!
    | edge_symbol
;

//==================================================================================================
pair_of_level_symbols
    : level_symbol level_symbol
    | IDENT
    | NUMBER
;

//==================================================================================================
level_symbol
    : IDENT
    | NUMBER
    | QMARK
    | MINUS
;

//==================================================================================================
edge_symbol
    : IDENT
    | STAR
;

//==================================================================================================
// A.6.1 Continuous Assignment Statements
//==================================================================================================

continuous_assign
    : KW_ASSIGN (strength)? (delay)? net_assignment (COMMA net_assignment)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CONTINUOUS_ASSIGNMENT
        ^(AST_STRENGTH (strength)?)
        ^(AST_DELAY    (delay)?)
        net_assignment // -> ^(AST_ASSIGNMENT ...)
    )+ // AST_CONTINUOUS_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
net_assignment
    : lvalue ASSIGN expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES lvalue)
        ^(AST_EXPRESSION expression)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.2 Procedural Blocks and Assignments
//==================================================================================================

//==================================================================================================
initial_construct
    : KW_INITIAL statement
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ACTIVITY
        ^(AST_ACTIVITY_TYPE KW_INITIAL)
        ^(AST_STATEMENT     statement)
    ) // AST_ACTIVITY
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
always_construct
    : KW_ALWAYS statement
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ACTIVITY
        ^(AST_ACTIVITY_TYPE KW_ALWAYS)
        ^(AST_STATEMENT     statement)
    ) // AST_ACTIVITY
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
discrete_assignment_statement
    : lvalue (type=ASSIGN | type=LESSEQ) control=delay_or_event_control expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGN_STATEMENT
        ^(AST_ASSIGN_TYPE   $type)
        ^(AST_EVENT_CONTROL ($control)?)
        ^(AST_ASSIGNMENT
            ^(AST_REFERENCES lvalue)
            ^(AST_EXPRESSION expression)
        ) // AST_ASSIGNMENT
    ) // AST_ASSIGN_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
delay_or_event_control
    : delay_control
    | event_control
    | repeat_event_control
    | /* Empty */
;

//==================================================================================================
continuous_assignment_statement
    : (
          type=KW_ASSIGN   assign=variable_assignment
        | type=KW_FORCE    assign=variable_assignment
        | type=KW_DEASSIGN assign=variable_deassignment
        | type=KW_RELEASE  assign=variable_deassignment
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGN_STATEMENT
        ^(AST_ASSIGN_TYPE   $type)
        ^(AST_EVENT_CONTROL /* null */)
        $assign             // -> ^(AST_ASSIGMENT ...)
    ) // AST_ASSIGN_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
variable_assignment
    : lvalue ASSIGN expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES lvalue)
        ^(AST_EXPRESSION expression)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
variable_deassignment
    : lvalue
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ASSIGNMENT
        ^(AST_REFERENCES lvalue)
        ^(AST_EXPRESSION /* null */)
    ) // AST_ASSIGNMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.3 Parallel and Sequential Blocks
//==================================================================================================

fork_block_statement
@init { boolean named = false; }
    : KW_FORK (COLON {named=true;} name=block_identifier
        (options {greedy=true;}: items+=block_item_declaration)*)?
        (statement)*
      KW_JOIN
//--------------------------------------------------------------------------------------------------
 -> {named}?
    ^(AST_BLOCK_STATEMENT
        $name             // -> ^(AST_NAME ...)
        ^(AST_BLOCK_TYPE  KW_FORK)
        ^(AST_BLOCK_ITEMS ($items)*)
        ^(AST_STATEMENTS  (statement)*)
    ) // AST_BLOCK_STATEMENT
 ->
    ^(AST_BLOCK_STATEMENT
        ^(AST_NAME        /* null */)
        ^(AST_BLOCK_TYPE  KW_FORK)
        ^(AST_BLOCK_ITEMS ($items)*)
        ^(AST_STATEMENTS  (statement)*)
    ) // AST_BLOCK_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
begin_block_statement
@init { boolean named = false; }
    : KW_BEGIN (COLON {named=true;} name=block_identifier
        (options {greedy=true;}: items+=block_item_declaration)*)?
        (statement)*
      KW_END
//--------------------------------------------------------------------------------------------------
 -> {named}?
    ^(AST_BLOCK_STATEMENT
        $name             // -> ^(AST_NAME ...)
        ^(AST_BLOCK_TYPE  KW_BEGIN)
        ^(AST_BLOCK_ITEMS ($items)*)
        ^(AST_STATEMENTS  (statement)*)
    ) // AST_BLOCK_STATEMENT
 ->
    ^(AST_BLOCK_STATEMENT
        ^(AST_NAME        /* null */)
        ^(AST_BLOCK_TYPE  KW_BEGIN)
        ^(AST_BLOCK_ITEMS ($items)*)
        ^(AST_STATEMENTS  (statement)*)
    ) // AST_BLOCK_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.4 Statements
//==================================================================================================

// NOTE: There is conflict between discrete_assignment_statement and task_enable,
// because they are both allowed to have the same prefix (hierarchical_identifier).
statement
    : attrs=attributes
      (
          (discrete_assignment_statement)=> item=discrete_assignment_statement   SEMI
        | (task_enable_statement)=>         item=task_enable_statement           SEMI
        |                                   item=continuous_assignment_statement SEMI
        |                                   item=system_task_enable_statement    SEMI
        |                                   item=event_trigger                   SEMI
        |                                   item=disable_statement               SEMI
        |                                   item=null_statement                  SEMI
        |                                   item=delayed_statement
        |                                   item=wait_statement
        |                                   item=if_statement
        |                                   item=case_statement
        |                                   item=loop_statement
        |                                   item=fork_block_statement
        |                                   item=begin_block_statement
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTED_STATEMENT
      $item
      ^(AST_ATTRIBUTES ($attrs)*)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
null_statement
    : /* null */
//--------------------------------------------------------------------------------------------------
 -> AST_NULL_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.5 Timing Control Statements
//==================================================================================================

delay_control
    : (
          SHARP delay_value
        | SHARP LPAREN mintypmax_expression RPAREN
    )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DELAY (delay_value)? (mintypmax_expression)?)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_control
    : AT name=hierarchical_event_identifier
//--------------------------------------------------------------------------------------------------
 -> ^(AST_EVENTS
        ^(AST_EVENT
            ^(AST_EDGE /* null */)
            ^(AST_EXPRESSION
                ^(AST_REFERENCE
                    $name // -> ^(AST_PATH ...)
                    ^(AST_SELECTION /* null */)
                ) // AST_REFERENCE
            ) // AST_EXRESSION
        ) // AST_EVENT
    ) // AST_EVENTS
//--------------------------------------------------------------------------------------------------
    | (
          AT LPAREN ( STAR | events=event_expression ) RPAREN
        | AT STAR
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_EVENTS ($events)?)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
repeat_event_control
    : KW_REPEAT LPAREN expression RPAREN event_control
//--------------------------------------------------------------------------------------------------
 -> ^(AST_REPEAT_EVENTS
        ^(AST_EXPRESSION expression)
        event_control // -> ^(AST_EVENTS ...)
    ) // AST_REPEAT_EVENTS
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_trigger
    : TRIGGER name=hierarchical_event_identifier (LBRACK expression RBRACK)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TRIGGER_STATEMENT
        $name // -> ^(AST_PATH ...)
        ^(AST_EXPRESSIONS (expression)*)
    ) // AST_TRIGGER_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_expression_1
    : (edge=KW_POSEDGE | edge=KW_NEGEDGE)? expr=expression
//--------------------------------------------------------------------------------------------------
 -> ^(AST_EVENT
        ^(AST_EDGE       ($edge)?)
        ^(AST_EXPRESSION $expr)
    ) // AST_EVENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_expression_2
    : event_expression_1 ((KW_OR | COMMA) event_expression_1)*
//--------------------------------------------------------------------------------------------------
 -> (event_expression_1)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
event_expression
    : event_expression_2
;

//==================================================================================================
procedural_timing_control
    : delay_control
    | event_control
;

//==================================================================================================
delayed_statement
    : control=procedural_timing_control body=statement
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DELAYED_STATEMENT
        ^(AST_EVENT_CONTROL $control)
        ^(AST_STATEMENT     $body)
    ) // AST_DELAYED_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
disable_statement
    : KW_DISABLE name=hierarchical_identifier
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DISABLE_STATEMENT
        $name // -> ^(AST_PATH ...)
    ) // AST_DISABLE_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
wait_statement
    : KW_WAIT LPAREN expr=expression RPAREN body=statement
//--------------------------------------------------------------------------------------------------
 -> ^(AST_WAIT_STATEMENT
        ^(AST_EXPRESSION $expr)
        ^(AST_STATEMENT  ($body)?)
    ) // AST_WAIT_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.6 Conditional Statements
//==================================================================================================

if_statement
    : KW_IF LPAREN expr=expression RPAREN
        then_body=statement
        ((KW_ELSE)=> KW_ELSE else_body=statement)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_IF_STATEMENT
        ^(AST_EXPRESSION  $expr)
        ^(AST_STATEMENT   $then_body)
        ^(AST_STATEMENT  ($else_body)?)
    ) // AST_IF_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.7 Case Statements
//==================================================================================================

case_statement
    : (type=KW_CASE | type=KW_CASEX | type=KW_CASEZ) LPAREN expr=expression RPAREN
        (case_item)+
      KW_ENDCASE
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CASE_STATEMENT
        ^(AST_CASE_TYPE  $type)
        ^(AST_EXPRESSION $expr)
        ^(AST_CASE_STATEMENT_ITEMS (case_item)+)
    ) // AST_CASE_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
case_item
    : (
        expr+=expression (COMMA expr+=expression)* COLON action=statement
      | KW_DEFAULT (COLON)? action=statement
    )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CASE_STATEMENT_ITEM
        ^(AST_EXPRESSIONS ($expr)*)
        ^(AST_STATEMENT   $action)
    ) // AST_CASE_STATEMENT_ITEM
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.8 Looping Statements
//==================================================================================================

loop_statement
    : (
          type=KW_FOREVER
            body=statement
        | type=KW_REPEAT LPAREN cond=expression RPAREN
            body=statement
        | type=KW_WHILE  LPAREN cond=expression RPAREN
            body=statement
        | type=KW_FOR LPAREN init=variable_assignment SEMI cond=expression SEMI step=variable_assignment RPAREN
            body=statement
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_LOOP_STATEMENT
        ^(AST_LOOP_TYPE  $type)
        ($init)?         // -> ^(AST_ASSIGNMENT ...) | null
        ^(AST_EXPRESSION ($cond)?)
        ($step)?         // -> ^(AST_ASSIGNMENT ...) | null
        ^(AST_STATEMENT  $body)
    ) // AST_LOOP_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.6.9 Task Enable Statements
//==================================================================================================

system_task_enable_statement
    : name=system_identifier (LPAREN (expression)? (COMMA (expression)?)* RPAREN)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TASK_STATEMENT
        ^(AST_PATH
            ^(AST_PATH_ITEM
                $name // -> ^(AST_NAME ...)
                ^(AST_EXPRESSION /* null */)
            ) // AST_PATH_ITEM
        ) // AST_PATH
        ^(AST_EXPRESSIONS (expression)*)
    ) // AST_TASK_STATEMENT
//--------------------------------------------------------------------------------------------------
;

task_enable_statement
    : name=hierarchical_task_identifier (LPAREN params=list_of_expressions RPAREN)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TASK_STATEMENT
        $name // -> ^(AST_PATH ...)
        ^(AST_EXPRESSIONS ($params)?)
    ) // AST_TASK_STATEMENT
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.7.* Specify Section
//==================================================================================================

specify_block
    : KW_SPECIFY
          (specify_item)*
      KW_ENDSPECIFY
//--------------------------------------------------------------------------------------------------
 -> ^(AST_SPECIFY (specify_item)*)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
specify_item
    : specparam_declaration SEMI!
    | pulsestyle_declaration
    | showcancelled_declaration
    | path_declaration
    | system_timing_check
;

//==================================================================================================
pulsestyle_declaration
    : type=(KW_PULSESTYLE_ONEVENT | KW_PULSESTYLE_ONDETECT)
        ports=list_of_port_references SEMI
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PULSESTYLE
        ^(AST_PULSESTYLE_TYPE $type)
        ^(AST_REFERENCES      $ports)
    ) // AST_PULSESTYLE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
showcancelled_declaration
    : type=(KW_SHOWCANCELLED | KW_NOSHOWCANCELLED)
        ports=list_of_port_references SEMI
//--------------------------------------------------------------------------------------------------
 -> ^(AST_SHOWCANCELLED
        ^(AST_SHOWCANCELLED_TYPE $type)
        ^(AST_REFERENCES         $ports)
    ) // AST_SHOWCANCELLED
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
path_declaration
    : state_independent_path_declaration SEMI!
    | state_dependent_path_declaration   SEMI!
;

//==================================================================================================
state_independent_path_declaration
    : path_description ASSIGN path_delay_value
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PATH_DECLARATION
        ^(AST_PATH_IF        /* null */)
        ^(AST_EXPRESSION     /* null */)
        ^(AST_PATH_REFERENCE path_description)
        path_delay_value     // -> ^(AST_DELAY ...)
        ^(AST_PATH_THEN      /* null */)
    ) // AST_PATH_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
path_description
    : LPAREN edge=(KW_POSEDGE | KW_NEGEDGE)?
        inputs=list_of_port_references op1=(PLUS | MINUS)?
          type=(PPATH | FPATH)
          (
              outputs=list_of_port_references
            | LPAREN
                outputs=list_of_port_references op2=(PLUS | MINUS)? COLON expr=expression
              RPAREN
          )
      RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PATH_DESCRIPTION
        ^(AST_PATH_TYPE   $type)
        ^(AST_EDGE       ($edge)?)
        ^(AST_REFERENCES  $inputs)
        ^(AST_OPERATION  ($op1)? ($op2)?)
        ^(AST_REFERENCES  $outputs)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_PATH_DESCRIPTION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_port_references
    : port_reference (options {greedy=true;}: COMMA port_reference)*
//--------------------------------------------------------------------------------------------------
 -> (port_reference)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
path_delay_value
    : (options {backtrack=true;}:
          list_of_mintypmax_expressions
        | LPAREN list_of_mintypmax_expressions RPAREN
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_DELAY list_of_mintypmax_expressions)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_mintypmax_expressions
    : mintypmax_expression (COMMA mintypmax_expression)*
//--------------------------------------------------------------------------------------------------
 -> (mintypmax_expression)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
state_dependent_path_declaration
    : (
          type=KW_IF LPAREN expr=expression RPAREN path=state_independent_path_declaration
        | type=KW_IFNONE                           path=state_independent_path_declaration
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PATH_DECLARATION
        ^(AST_PATH_IF         $type)
        ^(AST_EXPRESSION     ($expr)?)
        ^(AST_PATH_REFERENCE /* null */)
        ^(AST_DELAY          /* null */)
        ^(AST_PATH_THEN      $path)
    ) // AST_PATH_DECLARATION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
system_timing_check
    : setup_timing_check
    | hold_timing_check
    | setuphold_timing_check
    | recovery_timing_check
    | removal_timing_check
    | recrem_timing_check
    | skew_timing_check
    | timeskew_timing_check
    | fullskew_timing_check
    | period_timing_check
    | width_timing_check
    | nochange_timing_check
;

//==================================================================================================
setup_timing_check
    : SETUP_IDENT
        LPAREN
          data_event COMMA reference_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
hold_timing_check
    : HOLD_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
setuphold_timing_check
    : SETUPHOLD_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit COMMA timing_check_limit
          (
            COMMA (notifier)?
            (
              COMMA (stamptime_condition)?
              (
                COMMA (checktime_condition)?
                (
                  COMMA (delayed_reference)?
                  (
                    COMMA (delayed_data)?
                  )?
                )?
              )?
            )?
          )?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
recovery_timing_check
    : RECOVERY_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
removal_timing_check
    : REMOVAL_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
recrem_timing_check
    : RECREM_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit COMMA timing_check_limit
          (
            COMMA (notifier)?
            (
              COMMA (stamptime_condition)?
              (
                COMMA (checktime_condition)?
                (
                  COMMA (delayed_reference)?
                  (
                    COMMA (delayed_data)?
                  )?
                )?
              )?
            )?
          )?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
skew_timing_check
    : SKEW_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
timeskew_timing_check
    : TIMESKEW_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit
          (
            COMMA (notifier)?
            (
              COMMA (event_based_flag)?
              (
                COMMA (remain_active_flag)?
              )?
            )?
          )?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
fullskew_timing_check
    : FULLSKEW_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit COMMA timing_check_limit
          (
            COMMA (notifier)?
            (
              COMMA (event_based_flag)?
              (
                COMMA (remain_active_flag)?
              )?
            )?
          )?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
period_timing_check
    : PERIOD_IDENT
        LPAREN
          reference_event COMMA data_event COMMA timing_check_limit (COMMA notifier)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
width_timing_check
    : WIDTH_IDENT
        LPAREN
          reference_event COMMA timing_check_limit (COMMA threshold (COMMA notifier))?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
nochange_timing_check
    : NOCHANGE_IDENT
        LPAREN
          reference_event COMMA data_event COMMA start_edge_offset COMMA end_edge_offset
            (COMMA (notifier)?)?
        RPAREN SEMI
//--------------------------------------------------------------------------------------------------
 -> /* TODO: Ignored */
//--------------------------------------------------------------------------------------------------
;

data_event         : timing_check_event ;
reference_event    : timing_check_event ;
notifier           : variable_identifier ;
timing_check_limit : expression ;

timing_check_event
    : (timing_check_event_control)? specify_terminal_descriptor (AND3 timing_check_condition)?
;

timing_check_event_control
    : KW_POSEDGE
    | KW_NEGEDGE
    | edge_control_specifier
;

edge_control_specifier
  : KW_EDGE LBRACK edge_descriptor (COMMA edge_descriptor)* RBRACK
;

edge_descriptor
  : NUMBER (IDENT)?
  | IDENT
;

specify_terminal_descriptor : port_reference ;
timing_check_condition      : expression ;
stamptime_condition         : mintypmax_expression ;
checktime_condition         : mintypmax_expression ;
delayed_reference           : terminal_identifier (LBRACK mintypmax_expression RBRACK)? ;
delayed_data                : terminal_identifier (LBRACK mintypmax_expression RBRACK)? ;
event_based_flag            : expression ;
remain_active_flag          : expression ;
threshold                   : expression ;
start_edge_offset           : mintypmax_expression ;
end_edge_offset             : mintypmax_expression ;

//==================================================================================================
// A.8.1 Concatenations
//==================================================================================================

//==================================================================================================
concatenation
    : LCURLY list_of_expressions RCURLY
//--------------------------------------------------------------------------------------------------
 -> ^(AST_CONCATENATION list_of_expressions)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
replication
    : LCURLY expression concatenation RCURLY
//--------------------------------------------------------------------------------------------------
 -> ^(AST_REPLICATION
        ^(AST_EXPRESSION expression)
        concatenation // -> ^(AST_CONCATENATION ...)
    )
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.8.2 Function Calls
//==================================================================================================

//==================================================================================================
function_call
    : name=hierarchical_function_identifier attrs=attributes
      LPAREN params=list_of_expressions RPAREN
//--------------------------------------------------------------------------------------------------
 -> ^(AST_FUNCTION_CALL
        $name // -> ^(AST_PATH ...)
        ^(AST_EXPRESSIONS  $params)
        ^(AST_ATTRIBUTES  ($attrs)*)
    ) // AST_FUNCTION_CALL
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
system_function_call
    : name=system_identifier (LPAREN (expression)? (COMMA (expression)?)* RPAREN)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_FUNCTION_CALL
        ^(AST_PATH
            ^(AST_PATH_ITEM
                $name // -> ^(AST_NAME ...)
                ^(AST_EXPRESSION /* null */)
            ) // AST_PATH_ITEM
        ) // AST_PATH
        ^(AST_EXPRESSIONS (expression)*)
        ^(AST_ATTRIBUTES  /* null */)
    ) // AST_FUNCTION_CALL
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.8.3 Expressions
//==================================================================================================

//==================================================================================================
// Operator Priorities | Binary Operators
//==================================================================================================
//  ?:                                          | conditional_operator
//--------------------------------------------------------------------------------------------------
//  ||                                          | logical_or_operator
//--------------------------------------------------------------------------------------------------
//  &&                                          | logical_and_operator
//--------------------------------------------------------------------------------------------------
//  |       ~|                                  | bitwise_or_operator
//--------------------------------------------------------------------------------------------------
//  ^       ^~      ~^                          | bitwise_xor_operator
//--------------------------------------------------------------------------------------------------
//  &       ~&                                  | bitwise_and_operator
//--------------------------------------------------------------------------------------------------
//  ==      !=      ===     !=                  | equality_operator
//--------------------------------------------------------------------------------------------------
//  <       <=      >       >=                  | compare_operator
//--------------------------------------------------------------------------------------------------
//  <<      >>      <<<     >>>                 | shift_operator
//--------------------------------------------------------------------------------------------------
//  +       -                                   | additive_operator
//--------------------------------------------------------------------------------------------------
//  *       /       %                           | multiplicative_operator
//--------------------------------------------------------------------------------------------------
//  **                                          | power_operator
//==================================================================================================
// Operator Priorities | Unary Operators
//==================================================================================================
//  !                                           | unary_operator
//--------------------------------------------------------------------------------------------------
//  ~                                           | ...
//--------------------------------------------------------------------------------------------------
//  &                                           | ...
//--------------------------------------------------------------------------------------------------
//  ~&                                          | ...
//--------------------------------------------------------------------------------------------------
//  |                                           | ...
//--------------------------------------------------------------------------------------------------
//  ~|                                          | ...
//--------------------------------------------------------------------------------------------------
//  ^                                           | ...
//--------------------------------------------------------------------------------------------------
//  ~^      ^~                                  | ...
//--------------------------------------------------------------------------------------------------
//  +                                           | ...
//--------------------------------------------------------------------------------------------------
//  -                                           | ...
//==================================================================================================

//==================================================================================================
expression_0
    : primary
;

expression_1
    : expression_0
//    | LPAREN! expression RPAREN!
;

expression_2
    : expression_1
    | op=unary_operator attrs=attributes expression_1
//--------------------------------------------------------------------------------------------------
 -> ^(AST_UNARY_OPERATION ^(AST_OPERATION $op) expression_1 ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================

expression_3 : (expression_2 -> expression_2)
    (op=power_operator attrs=attributes rhs=expression_2
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_3 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_4 : (expression_3 -> expression_3)
    (op=multiplicative_operator attrs=attributes rhs=expression_3
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_4 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_5 : (expression_4 -> expression_4)
    (op=additive_operator attrs=attributes rhs=expression_4
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_5 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_6 : (expression_5 -> expression_5)
    (op=shift_operator attrs=attributes rhs=expression_5
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_6 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_7 : (expression_6 -> expression_6)
    (op=compare_operator attrs=attributes rhs=expression_6
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_7 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_8 : (expression_7 -> expression_7)
    (op=equality_operator attrs=attributes rhs=expression_7
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_8 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_9 : (expression_8 -> expression_8)
    (op=bitwise_and_operator attrs=attributes rhs=expression_8
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_9 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_10 : (expression_9 -> expression_9)
    (op=bitwise_xor_operator attrs=attributes rhs=expression_9
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_10 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_11 : (expression_10 -> expression_10)
    (op=bitwise_or_operator attrs=attributes rhs=expression_10
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_11 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_12 : (expression_11 -> expression_11)
    (op=logical_and_operator attrs=attributes rhs=expression_11
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_12 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

expression_13 : (expression_12 -> expression_12)
    (op=logical_or_operator attrs=attributes rhs=expression_12
//--------------------------------------------------------------------------------------------------
 -> ^(AST_BINARY_OPERATION ^(AST_OPERATION $op) $expression_13 $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)*;

//==================================================================================================
expression_14 : (cond=expression_13 -> expression_13)
    // Conditinal operator (in contrast to other ones) associates right to left
    (op=ternary_operator attrs=attributes lhs=expression_14 COLON rhs=expression_14
//--------------------------------------------------------------------------------------------------
 -> ^(AST_TERNARY_OPERATION ^(AST_OPERATION $op) $cond $lhs $rhs ^(AST_ATTRIBUTES ($attrs)*))
//--------------------------------------------------------------------------------------------------
)?;

//==================================================================================================
expression
    : expression_14
;

//==================================================================================================
mintypmax_expression
    : min=expression (COLON typ=expression COLON max=expression)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_MINTYPMAX_EXPRESSION
        ^(AST_EXPRESSION  $min)
        ^(AST_EXPRESSION ($typ)?)
        ^(AST_EXPRESSION ($max)?)
    ) // AST_MINTYPMAX_EXPRESSION
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
range_expression
    : lhs=expression ((type=PLUS COLON | type=MINUS COLON | type=COLON) rhs=expression)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_RANGE
        ^(AST_RANGE_TYPE ($type)?)
        ^(AST_EXPRESSION  $lhs)
        ^(AST_EXPRESSION ($rhs)?)
    ) // AST_RANGE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
list_of_expressions
    : expression (COMMA expression)*
//--------------------------------------------------------------------------------------------------
 -> (expression)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.8.4 Primaries
//==================================================================================================

primary
    : number
    | (function_call)=> function_call  // To resolve the ambiguity, the 'function_call' rule
    | reference                        // should precede the 'reference' rule
    | (replication)=> replication
    | concatenation
    | system_function_call
    | LPAREN! mintypmax_expression RPAREN!
    | string
;

//==================================================================================================
// A.8.5 Expression Left-Side Values
//==================================================================================================

lvalue
    : reference
//--------------------------------------------------------------------------------------------------
 -> reference
//--------------------------------------------------------------------------------------------------
    | LCURLY lvalue (COMMA lvalue)* RCURLY
//--------------------------------------------------------------------------------------------------
 -> (lvalue)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
reference
    : path=hierarchical_identifier ((LBRACK range_expression RBRACK)+)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_REFERENCE
        $path           // -> ^(AST_PATH ...)
        ^(AST_SELECTION (range_expression)*)
    ) // AST_REFERENCE
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.8.7 Numbers
//==================================================================================================

number
    : NUMBER
//--------------------------------------------------------------------------------------------------
 -> ^(AST_NUMBER NUMBER)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
string
    : STRING
//--------------------------------------------------------------------------------------------------
 -> ^(AST_STRING STRING)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// A.9.3 Identifiers
//==================================================================================================

identifier
    : (name=IDENT | name=ESCAPED_IDENT)
//--------------------------------------------------------------------------------------------------
 -> ^(AST_NAME $name)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
module_identifier          : identifier ;
udp_identifier             : identifier ;
block_identifier           : identifier ;
event_identifier           : identifier ;
function_identifier        : identifier ;
attribute_identifier       : identifier ;
generate_block_identifier  : identifier ;
genvar_identifier          : identifier ;
module_instance_identifier : identifier ;
net_identifier             : identifier ;
parameter_identifier       : identifier ;
port_identifier            : identifier ;
specparam_identifier       : identifier | name=PATHPULSE_PARAM -> ^(AST_NAME $name) ;
variable_identifier        : identifier ;
task_identifier            : identifier ;
terminal_identifier        : identifier ;

//==================================================================================================
hierarchical_identifier
    : hierarchical_identifier_item (DOT hierarchical_identifier_item)*
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PATH
      hierarchical_identifier_item+
    ) // AST_PATH
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
hierarchical_identifier_item
    : identifier
      (
          ((LBRACK expression RBRACK)? DOT)=> (LBRACK expression RBRACK)
        | /* No index */
      )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_PATH_ITEM
      identifier // -> ^(AST_NAME ...)
      ^(AST_EXPRESSION expression?)
    )+ // AST_PATH_ITEM
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
hierarchical_event_identifier     : hierarchical_identifier ;
hierarchical_function_identifier  : hierarchical_identifier ;
hierarchical_parameter_identifier : hierarchical_identifier ;
hierarchical_task_identifier      : hierarchical_identifier ;

//==================================================================================================
system_identifier
    : name=SYSTEM_IDENT
//--------------------------------------------------------------------------------------------------
 -> ^(AST_NAME $name)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
gate_identifier
    : ( name=KW_AND
      | name=KW_NAND
      | name=KW_NOR
      | name=KW_OR
      | name=KW_XOR
      | name=KW_XNOR
      | name=KW_BUF
      | name=KW_NOT
      | name=KW_BUFIF1
      | name=KW_BUFIF0
      | name=KW_NOTIF1
      | name=KW_NOTIF0
      | name=KW_CMOS
      | name=KW_NMOS
      | name=KW_PMOS
      | name=KW_RCMOS
      | name=KW_RNMOS
      | name=KW_RPMOS
      | name=KW_TRAN
      | name=KW_RTRAN
      | name=KW_TRANIF1
      | name=KW_RTRANIF1
      | name=KW_TRANIF0
      | name=KW_RTRANIF0
      | name=KW_PULLUP
      | name=KW_PULLDOWN )
//--------------------------------------------------------------------------------------------------
 -> ^(AST_NAME $name)
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
// Operators
//==================================================================================================
unary_operator
    : MINUS
    | PLUS
    | LNOT
    | BNOT
    | BAND
    | BNAND
    | BOR
    | BNOR
    | BXOR
    | BXNOR
;

//==================================================================================================
power_operator
    : POWER
;

//==================================================================================================
multiplicative_operator
    : STAR
    | DIV
    | MOD
;

//==================================================================================================
additive_operator
    : PLUS
    | MINUS
;

//==================================================================================================
shift_operator
    : LSR
    | LSL
    | ASR
    | ASL
;

//==================================================================================================
compare_operator
    : LESS
    | LESSEQ
    | GREATER
    | GREATEREQ
;

//==================================================================================================
equality_operator
    : EQ
    | NOTEQ
    | EQCASE
    | NOTEQCASE
;

//==================================================================================================
bitwise_and_operator
    : BAND
    | BNAND
;

//==================================================================================================
bitwise_xor_operator
    : BXOR
    | BXNOR
;

//==================================================================================================
bitwise_or_operator
    : BOR
    | BNOR
;

//==================================================================================================
logical_and_operator
    : LAND
;

//==================================================================================================
logical_or_operator
    : LOR
;

//==================================================================================================
ternary_operator
    : QMARK
;

//==================================================================================================
// Attributes
//==================================================================================================
attributes
    : (attribute_instance)*
;

//==================================================================================================
attribute_instance
    : LPAREN STAR attribute_assignment (COMMA attribute_assignment)* STAR RPAREN
//--------------------------------------------------------------------------------------------------
 -> (attribute_assignment)+
//--------------------------------------------------------------------------------------------------
;

//==================================================================================================
attribute_assignment
    : name=attribute_identifier (ASSIGN expr=expression)?
//--------------------------------------------------------------------------------------------------
 -> ^(AST_ATTRIBUTE
        $name            // -> ^(AST_NAME ...)
        ^(AST_EXPRESSION ($expr)?)
    ) // AST_ATTRIBUTE
//--------------------------------------------------------------------------------------------------
;
