/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.fortress.util.Pair;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.util.ModelUtils;
import ru.ispras.verilog.parser.transformer.VerilogTransformer;

import java.util.ArrayList;
import java.util.Collection;

/**
 * {@link VerilogCallCollector} collects information on function calls and
 * replaces function calls with newly created return-value variables.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCallCollector extends VerilogTransformer {
  private final Collection<Pair<VerilogDeclaration, VerilogExpression>> functionCalls = new ArrayList<>();

  public static Collection<Pair<VerilogDeclaration, VerilogExpression>> transform(final VerilogNode node) {
    InvariantChecks.checkNotNull(node);

    final VerilogCallCollector transformer = new VerilogCallCollector();
    transformer.run(node);

    return transformer.functionCalls;
  }

  @Override
  public Node transform(final Node node) {
    // The following rule replaces function calls to return-value variables.
    final TransformerRule callReplaceRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        // There should be operation.
        if (node.getKind() != Node.Kind.OPERATION) {
          return false;
        }

        // There should be call of a function.
        final NodeOperation operation = (NodeOperation) node;
        if (operation.getOperationId() != StandardOperation.FUNCTION) {
          return false;
        }

        final VerilogExpression functionCall = (VerilogExpression) operation.getUserData();
        InvariantChecks.checkNotNull(functionCall);

        final VerilogProcedure functionNode = functionCall.getDeclaration();
        InvariantChecks.checkNotNull(functionNode);

        // There should not be a system function.
        return !functionNode.isSystem();
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;

        final VerilogExpression functionCall = (VerilogExpression) operation.getUserData();
        InvariantChecks.checkNotNull(functionCall);

        final VerilogProcedure functionNode = functionCall.getDeclaration();
        InvariantChecks.checkNotNull(functionNode);

        // Generate a unique name for the return-value variable.
        final String returnPrefix = String.format("%s_call", functionNode.getName());
        final String returnName = VerilogElaborator.newVariableName(returnPrefix);

        // Create the return value declaration.
        final VerilogDeclaration returnDeclaration = new VerilogDeclaration(null);
        returnDeclaration.setName(returnName);
        returnDeclaration.setType(VerilogDeclaration.Type.VARIABLE);
        returnDeclaration.setElementType(functionNode.getElementType());
        returnDeclaration.setParent(functionNode.getParentNode());

        // Copying is necessary; otherwise, there will be a newly created variable node.
        functionCalls.add(new Pair<>(returnDeclaration, functionCall.clone()));

        return ModelUtils.getVariable(returnDeclaration);
      }
    };

    final NodeTransformer callReplaceTransformer = new NodeTransformer();
    callReplaceTransformer.addRule(StandardOperation.FUNCTION, callReplaceRule);

    final Node transformedNode = Transformer.transform(node, callReplaceTransformer);
    InvariantChecks.checkNotNull(transformedNode);

    return transformedNode;
  }
}
