/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.transformer.VerilogTransformerBvconcat;
import ru.ispras.verilog.parser.transformer.VerilogTransformerBvextract;
import ru.ispras.verilog.parser.transformer.VerilogTransformerComposite;
import ru.ispras.verilog.parser.transformer.VerilogTransformerOperation;
import ru.ispras.verilog.parser.transformer.VerilogTransformerVariableSubstitute;

import java.util.Map;

/**
 * {@link VerilogInstantiator} produces a process instance.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogInstantiator {
  private final VerilogProcess.Type type;
  private final VerilogNode node;
  private final VerilogContext context;
  private final boolean makeCopy;

  public VerilogInstantiator(
      final VerilogActivity activity,
      final VerilogContext context,
      final boolean makeCopy) {
    InvariantChecks.checkNotNull(activity);
    InvariantChecks.checkNotNull(context);

    // Process statement.
    this.type = activity.isAlways() ? VerilogProcess.Type.ALWAYS : VerilogProcess.Type.INITIAL;
    this.node = activity.getStatement();
    this.context = context;
    this.makeCopy = makeCopy;
  }

  public VerilogInstantiator(
      final VerilogAssign assign,
      final VerilogContext context,
      final boolean makeCopy) {
    InvariantChecks.checkNotNull(assign);
    InvariantChecks.checkNotNull(context);

    final VerilogAssignStatement statement = new VerilogAssignStatement(assign.getParentNode());

    // Continuous assignment statement.
    statement.setAssign();
    statement.setAssignment(assign.getAssignment());

    this.type = VerilogProcess.Type.ASSIGN;
    this.node = statement;
    this.context = context;
    this.makeCopy = makeCopy;
  }

  public VerilogInstantiator(
      final VerilogDeclaration declaration,
      final VerilogContext context,
      final boolean makeCopy) {
    InvariantChecks.checkNotNull(declaration);
    InvariantChecks.checkNotNull(context);

    // Continuous assignment statement.
    this.type = null;
    this.node = declaration;
    this.context = context;
    this.makeCopy = makeCopy;
  }

  public VerilogProcess instantiateProcess(
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    final VerilogStatement instance =
        (VerilogStatement) instantiate(node, variables, parameters);
    return new VerilogProcess(type, instance);
  }

  public VerilogDescriptor instantiateDescriptor(
      final NodeVariable variable,
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    final VerilogDeclaration instance =
        (VerilogDeclaration) instantiate(node, variables, parameters);
    return new VerilogDescriptor(variable, instance);
  }

  private VerilogNode instantiate(
      final VerilogNode node,
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(variables);
    InvariantChecks.checkNotNull(parameters);

    final VerilogNode instance = makeCopy ? node.clone() : node;
    final VerilogTransformerComposite transformer = new VerilogTransformerComposite();

    transformer.add(new VerilogTransformerVariableSubstitute(variables, context, parameters));
    transformer.add(new VerilogTransformerBvextract(variables));
    transformer.add(new VerilogTransformerBvconcat());
    transformer.add(new VerilogTransformerOperation());
    transformer.run(instance);

    return instance;
  }
}
