/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.interpreter;

import ru.ispras.fortress.calculator.ArityRange;
import ru.ispras.fortress.calculator.CalculatorEngine;
import ru.ispras.fortress.calculator.Operation;
import ru.ispras.fortress.calculator.OperationGroup;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.basis.VerilogLiteral;
import ru.ispras.verilog.parser.model.util.ModelUtils;

import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;

/**
 * {@link VerilogOperations} implements Verilog operations.
 * 
 * <p>
 * The main difference of {@link VerilogOperations} from the standard Fortress calculator is that
 * it takes into account X and Z masks stored in user data.
 * </p>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public enum VerilogOperations implements Operation<StandardOperation> {

  //================================================================================================
  // Unary Operations
  //================================================================================================

  PLUS(StandardOperation.PLUS, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);
      return operands[0];
    }
  },

  MINUS(StandardOperation.BVNEG, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();
      final BitVector xmask = literal.getXMask();
      final BitVector zmask = literal.getZMask();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.neg(value), xmask, zmask);
      return ModelUtils.newBitVector(result);
    }
  },

  NOT(StandardOperation.NOT, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final boolean value = literal.getBoolean();
      final VerilogLiteral result = new VerilogLiteral(!value);

      return ModelUtils.newBoolean(result);
    }
  },

  BVNOT(StandardOperation.BVNOT, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();
      final BitVector xmask = literal.getXMask();
      final BitVector zmask = literal.getZMask();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.not(value), xmask, zmask);
      return ModelUtils.newBitVector(result);
    }
  },

  //================================================================================================
  // Binary Operations
  //================================================================================================

  POWER(StandardOperation.POWER, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final int value2 = literal2.getInteger();
      InvariantChecks.checkTrue(value2 >= 0);

      BitVector result = BitVector.valueOf(1, value1.getBitSize());
      for (int i = 0; i < value2; i++) {
        result = BitVectorMath.mul(result, value1);
      }

      return ModelUtils.newBitVector(new VerilogLiteral(result));
    }
  },

  MUL(StandardOperation.BVMUL, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.mul(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  DIV(StandardOperation.BVSDIV, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.sdiv(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  MOD(StandardOperation.BVSMOD, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.smod(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  ADD(StandardOperation.BVADD, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.add(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  SUB(StandardOperation.BVSUB, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.sub(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  AND(StandardOperation.AND, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final boolean value1 = literal1.getBoolean();
      final boolean value2 = literal2.getBoolean();

      final VerilogLiteral result = new VerilogLiteral(value1 && value2);
      return ModelUtils.newBoolean(result);
    }
  },

  OR(StandardOperation.OR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final boolean value1 = literal1.getBoolean();
      final boolean value2 = literal2.getBoolean();

      final VerilogLiteral result = new VerilogLiteral(value1 || value2);
      return ModelUtils.newBoolean(result);
    }
  },

  LESS(StandardOperation.BVULT, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.ult(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  LESSEQ(StandardOperation.BVULE, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.ule(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  GREATER(StandardOperation.BVUGT, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.ugt(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  GREATEREQ(StandardOperation.BVUGE, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null && !literal1.hasXorZ());
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.uge(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  EQ(StandardOperation.EQ, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      // FIXME: X and Z.
      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.eq(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  NOTEQ(StandardOperation.NOTEQ, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();
      final int bitSize = Math.max(value1.getBitSize(), value2.getBitSize());

      // FIXME: X and Z.
      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.neq(value1.resize(bitSize, false), value2.resize(bitSize, false))
          );

      return ModelUtils.newBoolean(result);
    }
  },

  EQCASE(StandardOperation.EQCASE, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.eq(value1, value2));
      return ModelUtils.newBoolean(result);
    }
  },

  NOTEQCASE(StandardOperation.NOTEQCASE, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.neq(value1, value2));
      return ModelUtils.newBoolean(result);
    }
  },

  BVAND(StandardOperation.BVAND, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.and(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVNAND(StandardOperation.BVNAND, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.nand(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVXOR(StandardOperation.BVXOR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.xor(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVXNOR(StandardOperation.BVXNOR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.xnor(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVOR(StandardOperation.BVOR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.or(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVNOR(StandardOperation.BVNOR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal1);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkNotNull(literal2);

      final BitVector value1 = literal1.getBitVector();
      final BitVector value2 = literal2.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.nor(value1, value2));
      return ModelUtils.newBitVector(result);
    }
  },

  BVLSHR(StandardOperation.BVLSHR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector xmask1 = literal1.getXMask();
      final BitVector zmask1 = literal1.getZMask();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.lshr(value1, value2),
              xmask1 != null ? BitVectorMath.lshr(xmask1, value2) : null,
              zmask1 != null ? BitVectorMath.lshr(zmask1, value2) : null
          );

      return ModelUtils.newBitVector(result);
    }
  },

  BVLSHL(StandardOperation.BVLSHL, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector xmask1 = literal1.getXMask();
      final BitVector zmask1 = literal1.getZMask();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.shl(value1, value2),
              xmask1 != null ? BitVectorMath.shl(xmask1, value2) : null,
              zmask1 != null ? BitVectorMath.shl(zmask1, value2) : null
          );

      return ModelUtils.newBitVector(result);
    }
  },

  BVASHR(StandardOperation.BVASHR, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null && !literal2.hasXorZ());

      final BitVector value1 = literal1.getBitVector();
      final BitVector xmask1 = literal1.getXMask();
      final BitVector zmask1 = literal1.getZMask();
      final BitVector value2 = literal2.getBitVector();

      final VerilogLiteral result =
          new VerilogLiteral(
              BitVectorMath.ashr(value1, value2),
              xmask1 != null ? BitVectorMath.ashr(xmask1, value2) : null,
              zmask1 != null ? BitVectorMath.ashr(zmask1, value2) : null
          );

      return ModelUtils.newBitVector(result);
    }
  },

  BVASHL(StandardOperation.BVASHL, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      return BVLSHL.calculate(operands);
    }
  },

  BVCONCAT(StandardOperation.BVCONCAT, ArityRange.BINARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 2);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null);

      final BitVector value1 = literal1.getBitVector();
      final BitVector xmask1 = literal1.getXMask();
      final BitVector zmask1 = literal1.getZMask();

      final BitVector value2 = literal2.getBitVector();
      final BitVector xmask2 = literal2.getXMask();
      final BitVector zmask2 = literal2.getZMask();

      final BitVector value = BitVector.newMapping(value1, value2);

      final BitVector xmask = xmask1 != null || xmask2 != null
          ? BitVector.newMapping(
              xmask1 != null ? xmask1 : BitVector.newEmpty(value1.getBitSize()),
              xmask2 != null ? xmask2 : BitVector.newEmpty(value2.getBitSize()))
          : null;

      final BitVector zmask = zmask1 != null || zmask2 != null
          ? BitVector.newMapping(
              zmask1 != null ? zmask1 : BitVector.newEmpty(value1.getBitSize()),
              zmask2 != null ? zmask2 : BitVector.newEmpty(value2.getBitSize()))
          : null;

      final VerilogLiteral result = new VerilogLiteral(value, xmask, zmask);
      return ModelUtils.newBitVector(result);
    }
  },

  //================================================================================================
  // Ternary Operations
  //================================================================================================

  ITE(StandardOperation.ITE, ArityRange.TERNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 3);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final boolean value = literal.getBoolean();
      return value ? operands[1] : operands[2];
    }
  },

  BVEXTRACT(StandardOperation.BVEXTRACT, ArityRange.TERNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 3);

      final VerilogLiteral literal1 = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkTrue(literal1 != null);
      final VerilogLiteral literal2 = ModelUtils.extractLiteral(operands[1]);
      InvariantChecks.checkTrue(literal2 != null);
      final VerilogLiteral literal3 = ModelUtils.extractLiteral(operands[2]);
      InvariantChecks.checkTrue(literal3 != null);

      final int value1 = literal1.getBigInteger().intValue();
      final int value2 = literal2.getBigInteger().intValue();
      final BitVector value3 = literal3.getBitVector();
      final BitVector xmask3 = literal3.getXMask();
      final BitVector zmask3 = literal3.getZMask();

      final VerilogLiteral result =
          new VerilogLiteral(
              value3.field(value1, value2),
              xmask3 != null ? xmask3.field(value1, value2) : null,
              zmask3 != null ? zmask3.field(value1, value2) : null
          );

      return ModelUtils.newBitVector(result);
    }
  },

  //================================================================================================
  // Additional Unary Operations
  //================================================================================================

  BVANDR(StandardOperation.BVANDR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.andr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  BVNANDR(StandardOperation.BVNANDR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.nandr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  BVORR(StandardOperation.BVORR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.orr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  BVNORR(StandardOperation.BVNORR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.norr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  BVXORR(StandardOperation.BVXORR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.xorr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  BVXNORR(StandardOperation.BVXNORR, ArityRange.UNARY) {
    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length == 1);

      final VerilogLiteral literal = ModelUtils.extractLiteral(operands[0]);
      InvariantChecks.checkNotNull(literal);

      final BitVector value = literal.getBitVector();

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(BitVectorMath.xnorr(value));
      return ModelUtils.newBitVector(result);
    }
  },

  FUNCTION(StandardOperation.FUNCTION, ArityRange.UNARY_UNBOUNDED) {
    @Override
    public boolean validTypes(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length >= 1);

      final VerilogProcedure function = (VerilogProcedure) operands[0].getUserData();
      InvariantChecks.checkNotNull(function);

      return !function.isSystem();
    }

    @Override
    public Data calculate(final Data ... operands) {
      InvariantChecks.checkTrue(operands.length >= 1);

      final VerilogProcedure function = (VerilogProcedure) operands[0].getUserData();
      InvariantChecks.checkNotNull(function);

      VerilogLogger.debug("Function ports: %s", function.getPorts());

      final Map<String, Data> context = new HashMap<>();
      for (int i = 1; i < operands.length; i++) {
        final VerilogPort input = function.getPorts().get(i - 1);
        context.put(input.getName(), operands[i]);
      }

      VerilogLogger.debug("Evaluating %s(%s)", function.getName(), context);

      if (!VerilogInterpreter.interpret(function.getStatement(), context)) {
        VerilogLogger.debug("Cannot interpret the statement: %s", function.getStatement());
        return null;
      }

      final Data value = context.get(function.getOutput().getName());
      if (value == null) {
        VerilogLogger.debug("Cannot obtain the result of %s", function.getName());
        return null;
      }

      VerilogLogger.debug("Function result: %s (%s)", value, context);

      // FIXME: X and Z.
      final VerilogLiteral result = new VerilogLiteral(value);
      return ModelUtils.newBitVector(result);
    }
  };

  //================================================================================================
  // Calculator Engine
  //================================================================================================

  private static final CalculatorEngine CALCULATOR;

  static {
    final Map<StandardOperation, Operation<StandardOperation>> operations;
    operations = new EnumMap<>(StandardOperation.class);

    for (Operation<StandardOperation> value : values()) {
      operations.put(value.getOperationId(), value);
    }

    final OperationGroup<StandardOperation> standardOperations = new OperationGroup<>();
    standardOperations.registerOperations(DataTypeId.BIT_VECTOR, operations);
    standardOperations.registerOperations(DataTypeId.LOGIC_INTEGER, operations);
    standardOperations.registerOperations(DataTypeId.LOGIC_BOOLEAN, operations);
    standardOperations.registerOperations(DataTypeId.LOGIC_REAL, operations);
    standardOperations.registerOperations(DataTypeId.UNKNOWN, operations);

    CALCULATOR = standardOperations;
  }

  public static CalculatorEngine getCalculatorEngine() {
    return CALCULATOR;
  }

  //================================================================================================
  // Operation Interface Implementation
  //================================================================================================

  private final StandardOperation operationId;
  private final ArityRange operationArity;

  private VerilogOperations(final StandardOperation operationId, final ArityRange operationArity) {
    InvariantChecks.checkNotNull(operationId);
    InvariantChecks.checkNotNull(operationArity);

    this.operationId = operationId;
    this.operationArity = operationArity;
  }

  @Override
  public final StandardOperation getOperationId() {
    return operationId;
  }

  @Override
  public final ArityRange getOperationArity() {
    return operationArity;
  }

  @Override
  public boolean validTypes(final Data ... operands) {
    return true;
  }
}
