/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogStrength;

import java.util.EnumSet;

/**
 * {@code VerilogAssign} represents continuous assignments ({@code assign}).
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogAssign extends VerilogNode {
  public static final Tag TAG = Tag.ASSIGN;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Drive strength. */
  private VerilogStrength strength;
  /** Assignment delay. */
  private VerilogDelay delay;
  /** Assignment (a pair of l-value and expression). */
  private VerilogAssignment assignment;

  /**
   * Constructs a continuous assignment.
   * 
   * @param parent the parent node.
   */
  public VerilogAssign(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the continuous assignment.
   * 
   * @param other the continuous assignment to be copied.
   * @param parent the parent node.
   */
  public VerilogAssign(final VerilogAssign other, final VerilogNode parent) {
    super(other, parent);

    this.strength = other.strength != null ? other.strength.clone() : null;
    this.delay = other.delay != null ? other.delay.clone() : null;
    this.assignment = other.assignment != null ? other.assignment.clone() : null;
  }

  /**
   * Returns the drive strength.
   * 
   * @return the drive strength.
   */
  public VerilogStrength getStrength() {
    return strength;
  }

  /**
   * Sets the drive strength.
   * 
   * @param strength the strength to be set.
   */
  public void setStrength(final VerilogStrength strength) {
    InvariantChecks.checkNotNull(strength);
    this.strength = strength;
  }

  /**
   * Returns the assignment delay.
   * 
   * @return the assignment delay.
   */
  public VerilogDelay getDelay() {
    return delay;
  }

  /**
   * Sets the assignment delay.
   * 
   * @param delay the delay to be set.
   */
  public void setDelay(final VerilogDelay delay) {
    InvariantChecks.checkNotNull(delay);
    this.delay = delay;
  }

  /**
   * Returns the assignment itself.
   * 
   * @return the assignment.
   */
  public VerilogAssignment getAssignment() {
    return assignment;
  }

  /**
   * Sets the assignment.
   * 
   * @param assignment the assignment to be set.
   */
  public void setAssignment(final VerilogAssignment assignment) {
    InvariantChecks.checkNotNull(assignment);
    this.assignment = assignment;
  }

  @Override
  public VerilogAssign clone() {
    return new VerilogAssign(this, getParentNode());
  }
}
