/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.EnumSet;

/**
 * {@link VerilogIfGenerate} represents if-then-else generate constructs.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogIfGenerate extends VerilogNode {
  public static final Tag TAG = Tag.IF_GENERATE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.IF_GENERATE_BRANCH);

  /** Condition. */
  private VerilogExpression expression;

  /**
   * Creates a if-then-else generate construct.
   * 
   * @param parent the parent node.
   */
  public VerilogIfGenerate(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.PSEUDO_ELEMENT, parent);
  }

  /**
   * Creates a if-then-else generate construct.
   * 
   * @param other the if-then-else generate construct to be copied.
   * @param parent the parent node.
   */
  public VerilogIfGenerate(final VerilogIfGenerate other, final VerilogNode parent) {
    super(other, parent);

    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the condition.
   * 
   * @return the condition.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the condition.
   * 
   * @param expression the condition.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the generate block related to the branch {@code then}.
   * 
   * @return the {@code then}'s generate block.
   */
  public VerilogBlockGenerate getThenGenerate() {
    for (final AbstractNode<VerilogNode.Tag> item : items(TAGS_CHILDREN)) {
      final VerilogIfGenerateBranch branch = (VerilogIfGenerateBranch) item;

      if (branch.isThen()) {
        return branch.getGenerate();
      }
    }

    return null;
  }

  /**
   * Sets the generate block related to the branch {@code then}.
   * 
   * @param thenGenerate the generate block to be set.
   */
  public void setThenGenerate(final VerilogBlockGenerate thenGenerate) {
    final VerilogIfGenerateBranch branch =
        new VerilogIfGenerateBranch(VerilogIfGenerateBranch.Type.THEN, thenGenerate, this);

    add(branch);
  }

  /**
   * Returns the generate block related to the branch {@code else}.
   * 
   * @return the {@code else}'s generate block.
   */
  public VerilogBlockGenerate getElseGenerate() {
    for (final AbstractNode<VerilogNode.Tag> item : items(TAGS_CHILDREN)) {
      final VerilogIfGenerateBranch branch = (VerilogIfGenerateBranch) item;

      if (branch.isElse()) {
        return branch.getGenerate();
      }
    }

    return null;
  }

  /**
   * Sets the generate block related to the else branch.
   * 
   * @param elseGenerate the generate block to be set.
   */
  public void setElseGenerate(final VerilogBlockGenerate elseGenerate) {
    final VerilogIfGenerateBranch branch =
        new VerilogIfGenerateBranch(VerilogIfGenerateBranch.Type.ELSE, elseGenerate, this);

    add(branch);
  }

  @Override
  public VerilogIfGenerate clone() {
    return new VerilogIfGenerate(this, getParentNode());
  }
}
