/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.EnumSet;

/**
 * {@link VerilogLoopGenerate} represents generate loops.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogLoopGenerate extends VerilogNode {
  public static final Tag TAG = Tag.LOOP_GENERATE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.BLOCK_GENERATE);

  /** Loop initialization assignment. */
  private VerilogAssignment initialization;
  /** Loop continue condition. */
  private VerilogExpression expression;
  /** Loop iteration assignment. */
  private VerilogAssignment iteration;

  /**
   * Creates a generate loop.
   * 
   * @param parent the parent node.
   */
  public VerilogLoopGenerate(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.PSEUDO_ELEMENT, parent);
  }

  /**
   * Creates a copy of the generate loop.
   * 
   * @param other the generate loop to be copied.
   * @param parent the parent node.
   */
  public VerilogLoopGenerate(final VerilogLoopGenerate other, final VerilogNode parent) {
    super(other, parent);

    this.initialization = other.initialization != null ? other.initialization.clone() : null;
    this.expression = other.expression != null ? other.expression.clone() : null;
    this.iteration = other.iteration != null ? other.iteration.clone() : null;
  }

  /**
   * Returns the initialization assignment of the loop.
   * 
   * @return the initialization assignment.
   */
  public VerilogAssignment getInitialization() {
    return initialization;
  }

  /**
   * Sets the initialization assignment of the loop.
   * 
   * @param initialization the assignment to be set.
   */
  public void setInitialization(final VerilogAssignment initialization) {
    InvariantChecks.checkNotNull(initialization);
    this.initialization = initialization;
  }

  /**
   * Returns the continue condition of the loop.
   * 
   * @return the condition.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the continue condition of the loop.
   * 
   * @param expression the condition to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the iteration assignment of the loop.
   * 
   * @return the iteration assignment.
   */
  public VerilogAssignment getIteration() {
    return iteration;
  }

  /**
   * Sets the iteration assignment of the loop.
   * 
   * @param iteration the assignment to be set.
   */
  public void setIteration(final VerilogAssignment iteration) {
    InvariantChecks.checkNotNull(iteration);
    this.iteration = iteration;
  }

  /**
   * Returns the loop body.
   * 
   * @return the loop body.
   */
  public VerilogBlockGenerate getGenerate() {
    return (VerilogBlockGenerate) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the loop body.
   * 
   * @param generate the body to be set.
   */
  public void setGenerate(final VerilogBlockGenerate generate) {
    add(generate);
  }

  @Override
  public VerilogLoopGenerate clone() {
    return new VerilogLoopGenerate(this, getParentNode());
  }
}
