/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogDelay} represents the abstract syntax of the delay.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogDelay {
  /** List of expressions representing delay values. */
  private List<VerilogMinTypMax> delays = new ArrayList<>();

  /**
   * Creates an empty delay object.
   */
  public VerilogDelay() {}

  /**
   * Creates a copy of the delay object.
   * 
   * @param other the delay object to be copied.
   */
  public VerilogDelay(final VerilogDelay other) {
    InvariantChecks.checkNotNull(other);

    this.delays = new ArrayList<>(other.delays.size());
    for (final VerilogMinTypMax minTypMax : other.delays) {
      this.delays.add(minTypMax.clone());
    }
  }

  /**
   * Checks whether the delay is specified (i.e. the list of delays is not empty).
   * 
   * @return {@code true} iff the delay is specified.
   */
  public boolean isSpecified() {
    return !delays.isEmpty();
  }

  /**
   * Returns the list of expressions representing delay values.
   * 
   * @return the list of delays.
   */
  public List<VerilogMinTypMax> getDelays() {
    return delays;
  }

  /**
   * Adds a delay value to the list.
   * 
   * @param delay the delay value to be added.
   */
  public void addDelay(final VerilogMinTypMax delay) {
    delays.add(delay);
  }

  @Override
  public String toString() {
    return delays.toString();
  }

  @Override
  public VerilogDelay clone() {
    return new VerilogDelay(this);
  }
}
