/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogElementType} represents element types.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogElementType {

  /** {@link Type} contains the element types. */
  public enum Type {
    /** Undefined-type variable: {@code x}. */
    UNDEFINED,
    /** INTEGER variable: {@code integer x}. */
    INTEGER,
    /** REAL variable: {@code real x}. */
    REAL,
    /** TIME variable: {@code time x}. */
    TIME,
    /** REALTIME variable: {@code realtime x}. */
    REALTIME,
    /** SUPPLY0 net: {@code supply0 x}. */
    SUPPLY0,
    /** SUPPLY1 net: {@code supply1 x}. */
    SUPPLY1,
    /** TRI net: {@code tri x}. */
    TRI,
    /** TRIAND net: {@code triand x}. */
    TRIAND,
    /** TRIOR net: {@code trior x}. */
    TRIOR,
    /** TRI0 net: {@code tri0 x}. */
    TRI0,
    /** TRI1 net: {@code tri1 x}. */
    TRI1,
    /** UWIRE net: {@code uwire x}. */
    UWIRE,
    /** WIRE net: {@code wire x}. */
    WIRE,
    /** WAND net: {@code wand x}. */
    WAND,
    /** WOR net: {@code wor x}. */
    WOR,
    /** TRIREG: {@code trireg x}. */
    TRIREG,
    /** REG: {@code reg x}. */
    REG
  }

  /** By default, the type is undefined. */
  private Type type = Type.UNDEFINED;
  /** Drive and charge strength. */
  private VerilogStrength strength;

  /** Vectored element. */
  private boolean vectored = false;
  /** Scalared element. */
  private boolean scalared = false;
  /** Signed-type element. */
  private boolean signed = false;

  /** Element range. */
  private VerilogRange range;

  /**
   * Creates an element type.
   */
  public VerilogElementType() {}

  /**
   * Creates a copy of the element type.
   * 
   * @param other the element type to be copied.
   */
  public VerilogElementType(final VerilogElementType other) {
    InvariantChecks.checkNotNull(other);

    this.type = other.type;
    this.vectored = other.vectored;
    this.scalared = other.scalared;
    this.signed = other.signed;
    this.range = other.range != null ? other.range.clone() : null;
  }

  /**
   * Checks whether the type is specified (something is written in the code).
   * 
   * @return {@code true} iff the type is specified.
   */
  public boolean isSpecified() {
    return type != Type.UNDEFINED || vectored || scalared || signed || range != null;
  }

  /**
   * Returns the element type.
   *
   * @return the element type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the element type.
   *
   * @param type the element type to be set.
   */
  public void setType(final Type type) {
    this.type = type;
  }

  /**
   * Checks whether the element is {@code INTEGER}.
   *
   * @return {@code true} iff the element is {@code INTEGER}.
   */
  public boolean isInteger() {
    return type == Type.INTEGER;
  }

  /**
   * Checks whether the element is {@code REAL}.
   *
   * @return {@code true} iff the element is {@code REAL}.
   */
  public boolean isReal() {
    return type == Type.REAL;
  }

  /**
   * Checks whether the element is {@code TIME}.
   *
   * @return {@code true} iff the element is {@code TIME}.
   */
  public boolean isTime() {
    return type == Type.TIME;
  }

  /**
   * Checks whether the element is {@code REALTIME}.
   *
   * @return {@code true} iff the element is {@code REALTIME}.
   */
  public boolean isRealtime() {
    return type == Type.REALTIME;
  }

  /**
   * Checks whether the element is {@code SUPPLY0}.
   *
   * @return {@code true} iff the element is {@code SUPPLY0}.
   */
  public boolean isSupply0() {
    return type == Type.SUPPLY0;
  }

  /**
   * Checks whether the element is {@code SUPPLY1}.
   *
   * @return {@code true} iff the element is {@code SUPPLY1}.
   */
  public boolean isSupply1() {
    return type == Type.SUPPLY1;
  }

  /**
   * Checks whether the element is {@code TRI}.
   *
   * @return {@code true} iff the element is {@code TRI}.
   */
  public boolean isTri() {
    return type == Type.TRI;
  }

  /**
   * Checks whether the element is {@code TRIAND}.
   *
   * @return {@code true} iff the element is {@code TRIAND}.
   */
  public boolean isTriand() {
    return type == Type.TRIAND;
  }

  /**
   * Checks whether the element is {@code TRIOR}.
   *
   * @return {@code true} iff the element is {@code TRIOR}.
   */
  public boolean isTrior() {
    return type == Type.TRIOR;
  }

  /**
   * Checks whether the element is {@code TRI0}.
   *
   * @return {@code true} iff the element is {@code TRI0}.
   */
  public boolean isTri0() {
    return type == Type.TRI0;
  }

  /**
   * Checks whether the element is {@code TRI1}.
   *
   * @return {@code true} iff the element is {@code TRI1}.
   */
  public boolean isTri1() {
    return type == Type.TRI1;
  }

  /**
   * Checks whether the element is {@code UWIRE}.
   *
   * @return {@code true} iff the element is {@code UWIRE}.
   */
  public boolean isUwire() {
    return type == Type.UWIRE;
  }

  /**
   * Checks whether the element is {@code WIRE}.
   *
   * @return {@code true} iff the element is {@code WIRE}.
   */
  public boolean isWire() {
    return type == Type.WIRE;
  }

  /**
   * Checks whether the element is {@code WAND}.
   *
   * @return {@code true} iff the element is {@code WAND}.
   */
  public boolean isWand() {
    return type == Type.WAND;
  }

  /**
   * Checks whether the element is {@code WOR}.
   *
   * @return {@code true} iff the element {@code WOR}.
   */
  public boolean isWor() {
    return type == Type.WOR;
  }

  /**
   * Checks whether the element is {@code TRIREG}.
   *
   * @return {@code true} iff the element {@code TRIREG}.
   */
  public boolean isTrireg() {
    return type == Type.TRIREG;
  }

  /**
   * Checks whether the element is {@code REG}.
   *
   * @return {@code true} iff the element is {@code REG}.
   */
  public boolean isReg() {
    return type == Type.REG;
  }

  /**
   * Checks whether the element is a numeric variable ({@code INTEGER} or {@code REAL}).
   *
   * @return {@code true} iff the element is a numeric variable.
   */
  public boolean isNumeric() {
    return isInteger() || isReal();
  }

  /**
   * Checks whether the element is a temporal variable ({@code TIME} or {@code REALTIME}).
   *
   * @return {@code true} iff the element is a temporal variable.
   */
  public boolean isTemporal() {
    return isTime() || isRealtime();
  }

  /**
   * Checks whether the element is a register.
   *
   * @return {@code true} iff the element is a register.
   */
  public boolean isRegister() {
    return isReg();
  }

  /**
   * Checks whether the element is a net.
   *
   * @return {@code true} iff the element is a net.
   */
  public boolean isNet() {
    return !isNumeric() && !isTemporal() && !isRegister();
  }

  /**
   * Sets the element type to {@code INTEGER}.
   */
  public void setInteger() {
    type = Type.INTEGER;
  }

  /**
   * Sets the element type to {@code REAL}.
   */
  public void setReal() {
    type = Type.REAL;
  }

  /**
   * Sets the element type to {@code TIME}.
   */
  public void setTime() {
    type = Type.TIME;
  }

  /**
   * Sets the element type to {@code REALTIME}.
   */
  public void setRealtime() {
    type = Type.REALTIME;
  }

  /**
   * Sets the element type to {@code SUPPLY0}.
   */
  public void setSupply0() {
    type = Type.SUPPLY0;
  }

  /**
   * Sets the element type to {@code SUPPLY1}.
   */
  public void setSupply1() {
    type = Type.SUPPLY1;
  }

  /**
   * Sets the element type to {@code TRI}.
   */
  public void setTri() {
    type = Type.TRI;
  }

  /**
   * Sets the element type to {@code TRIAND}.
   */
  public void setTriand() {
    type = Type.TRIAND;
  }

  /**
   * Sets the element type to {@code TRIOR}.
   */
  public void setTrior() {
    type = Type.TRIOR;
  }

  /**
   * Sets the element type to {@code TRI0}.
   */
  public void setTri0() {
    type = Type.TRI0;
  }

  /**
   * Sets the element type to {@code TRI1}.
   */
  public void setTri1() {
    type = Type.TRI1;
  }

  /**
   * Sets the element type to {@code UWIRE}.
   */
  public void setUwire() {
    type = Type.UWIRE;
  }

  /**
   * Sets the element type to {@code WIRE}.
   */
  public void setWire() {
    type = Type.WIRE;
  }

  /**
   * Sets the element type to {@code WAND}.
   */
  public void setWand() {
    type = Type.WAND;
  }

  /**
   * Sets the element type to {@code WOR}.
   */
  public void setWor() {
    type = Type.WOR;
  }

  /**
   * Sets the element type to {@code TRIREG}.
   */
  public void setTrireg() {
    type = Type.TRIREG;
  }

  /**
   * Sets the element type to {@code REG}.
   */
  public void setReg() {
    type = Type.REG;
  }

  /**
   * Returns the strength.
   * 
   * @return the strength.
   */
  public VerilogStrength getStrength() {
    return strength;
  }

  /**
   * Sets the strength.
   * 
   * @param strength the strength to be set.
   */
  public void setStrength(final VerilogStrength strength) {
    InvariantChecks.checkNotNull(strength);
    this.strength = strength;
  }

  /**
   * Checks whether the element is vectored.
   *
   * @return {@code true} iff the element is vectored.
   */
  public boolean isVectored() {
    return vectored;
  }

  /**
   * Sets the vectored property.
   */
  public void setVectored() {
    vectored = true;
  }

  /**
   * Checks whether the element is scalared.
   *
   * @return {@code true} iff the element is scalared.
   */
  public boolean isScalared() {
    return scalared;
  }

  /**
   * Sets the scalared property.
   */
  public void setScalared() {
    scalared = true;
  }

  /**
   * Checks whether the element is signed.
   *
   * @return {@code true} iff the element is signed.
   */
  public boolean isSigned() {
    return signed;
  }

  /**
   * Sets the signed property.
   */
  public void setSigned() {
    signed = true;
  }

  /**
   * Returns the range of the element.
   *
   * @return the element range.
   */
  public VerilogRange getRange() {
    return range;
  }

  /**
   * Sets a range of the element.
   *
   * @param range the element range.
   */
  public void setRange(final VerilogRange range) {
    InvariantChecks.checkNotNull(range);
    this.range = range;
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();

    builder.append(type);

    builder.append(vectored ? " vectored" : "");
    builder.append(scalared ? " scalared" : "");
    builder.append(signed   ? " signed"   : "");

    if (range != null) {
      builder.append(" ");
      builder.append(range);
    }

    return builder.toString();
  }

  @Override
  public VerilogElementType clone() {
    return new VerilogElementType(this);
  }
}
