/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogPathDescription} represents path descriptions.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogPathDescription {

  /** {@link Type} contains the path description types. */
  public enum Type {
    /** Parallel path description: {@code source => destination}. */
    PARALLEL,
    /** Full path description: {@code source *> destination}. */
    FULL
  }

  /** {@link Polarity} contains the polarity types. */
  public enum Polarity {
    /** Undefined polarity. */
    UNDEFINED,
    /** Positive polarity. */
    PLUS,
    /** Negative polarity. */
    MINUS
  }

  /** Path description type. */
  private Type type;
  /** Edge type (only for edge sensitive path descriptions). */
  private VerilogEdge edge;
  /** By default, the polarity is undefined. */
  private Polarity polarity = Polarity.UNDEFINED;

  /** Sources. */
  private List<VerilogReference> inputs = new ArrayList<>();
  /** Destinations. */
  private List<VerilogReference> outputs = new ArrayList<>();
  /** Data source expression. */
  private VerilogExpression expression;

  /**
   * Creates a path description.
   */
  public VerilogPathDescription() {}

  /**
   * Creates a copy of the path description.
   * 
   * @param other the path description to be copied.
   */
  public VerilogPathDescription(final VerilogPathDescription other) {
    InvariantChecks.checkNotNull(other);

    this.type = other.type;
    this.edge = other.edge;
    this.polarity = other.polarity;

    this.inputs = new ArrayList<>(other.inputs.size());
    for (final VerilogReference reference : other.inputs) {
      this.inputs.add(reference.clone());
    }

    this.outputs = new ArrayList<>(other.outputs.size());
    for (final VerilogReference reference : other.outputs) {
      this.outputs.add(reference.clone());
    }

    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the type of the path description.
   * 
   * @return the path description type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the path description.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the path description is {@code PARALLEL}.
   * 
   * @return {@code true} iff the path description {@code PARALLEL}.
   */
  public boolean isParallel() {
    return type == Type.PARALLEL;
  }

  /**
   * Checks whether the path description is {@code FULL}.
   * 
   * @return {@code true} iff the path description is {@code FULL}.
   */
  public boolean isFull() {
    return type == Type.FULL;
  }

  /**
   * Sets the path description type to {@code PARALLEL}.
   */
  public void setParallel() {
    type = Type.PARALLEL;
  }

  /**
   * Sets the path description type to {@code FULL}.
   */
  public void setFull() {
    type = Type.FULL;
  }

  /**
   * Returns the polarity of the path description.
   * 
   * @return the path description polarity.
   */
  public Polarity getPolarity() {
    return polarity;
  }

  /**
   * Sets the polarity of the path description.
   * 
   * @param polarity the polarity to be set.
   */
  public void setPolarity(final Polarity polarity) {
    InvariantChecks.checkNotNull(polarity);
    this.polarity = polarity;
  }

  /**
   * Checks whether the polarity is {@code PLUS}.
   * 
   * @return {@code true} iff the polarity is {@code PLUS}.
   */
  public boolean isPlus() {
    return polarity == Polarity.PLUS;
  }

  /**
   * Checks whether the polarity is {@code MINUS}.
   * 
   * @return {@code true} if the polarity is {@code MINUS}.
   */
  public boolean isMinus() {
    return polarity == Polarity.MINUS;
  }

  /**
   * Sets the path description polarity to {@code PLUS}.
   */
  public void setPlus() {
    polarity = Polarity.PLUS;
  }

  /**
   * Sets the path description polarity to {@code MINUS}.
   */
  public void setMinus() {
    polarity = Polarity.MINUS;
  }

  /**
   * Returns the edge type of the path description.
   * 
   * @return the path description edge.
   */
  public VerilogEdge getEdge() {
    return edge;
  }

  /**
   * Sets the edge type of the path description.
   * 
   * @param edge the edge type to be set.
   */
  public void setEdge(final VerilogEdge edge) {
    InvariantChecks.checkNotNull(edge);
    this.edge = edge;
  }

  /**
   * Returns the source references of the path description.
   * 
   * @return the source references.
   */
  public List<VerilogReference> getInputs() {
    return inputs;
  }

  /**
   * Adds the reference to the source references.
   * 
   * @param input the reference to be added.
   */
  public void addInput(final VerilogReference input) {
    InvariantChecks.checkNotNull(input);
    inputs.add(input);
  }

  /**
   * Returns the destination references of the path description.
   * 
   * @return the destination references.
   */
  public List<VerilogReference> getOutputs() {
    return outputs;
  }

  /**
   * Adds the reference to the destination references.
   * 
   * @param output the reference to be added.
   */
  public void addOutput(final VerilogReference output) {
    InvariantChecks.checkNotNull(output);
    outputs.add(output);
  }

  /**
   * Returns the data source expression of the path description.
   * 
   * @return the data source expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the data source expression of the path description.
   * 
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  @Override
  public String toString() {
    return String.format("%s:%s:%s", type, edge, polarity);
  }

  @Override
  public VerilogPathDescription clone() {
    return new VerilogPathDescription(this);
  }
}
