/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.util;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.ValueProvider;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.elaborator.VerilogIndexRange;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogLiteral;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogRange;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class ModelUtils {

  private ModelUtils() {}

  //================================================================================================
  // Utilities for VerilogLiteral
  //================================================================================================

  public static NodeValue getValue(final VerilogLiteral literal) {
    final Data data;

    switch (literal.getType()) {
      case INTEGER:
        data = newInteger(literal);
        break;
      case REAL:
        data = newReal(literal);
        break;
      default:
        data = newBitVector(literal);
        break;
    }

    // Store some Verilog-specific information (e.g., X- and Z-masks).
    data.setUserData(literal);
    return new NodeValue(data);
  }

  public static VerilogLiteral extractLiteral(final Data data) {
    final VerilogLiteral literal = (VerilogLiteral) data.getUserData();
    return literal != null ? literal : new VerilogLiteral(data);
  }

  public static Data newBitVector(final VerilogLiteral literal) {
    final Data data = Data.newBitVector(literal.getBitVector());
    data.setUserData(literal);

    return data;
  }

  public static Data newInteger(final VerilogLiteral literal) {
    final Data data = Data.newInteger(literal.getInteger());
    data.setUserData(literal);

    return data;
  }

  public static Data newBoolean(final VerilogLiteral literal) {
    final Data data = Data.newBoolean(literal.getBoolean());
    data.setUserData(literal);

    return data;
  }

  public static Data newReal(final VerilogLiteral literal) {
    final Data data = Data.newReal(literal.getReal());
    data.setUserData(literal);

    return data;
  }

  //================================================================================================
  // Utilities for VerilogReference
  //================================================================================================

  public static NodeVariable getVariable(
      final VerilogReference reference,
      final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(reference);

    // The declaration has been set by the static checker.
    final VerilogDeclaration declaration = reference.getDeclaration();

    final String name = reference.getPath().toString();

    // NOTE: The declaration may differ from VerilogDeclaration (i.e. a variable declaration).
    // It may be VerilogModule (e.g., reference from the $dumpvars system task).
    final DataType type = declaration != null
        ? getDataType(declaration, valueProvider, null)
        : DataType.UNKNOWN;

    final Variable variable = new Variable(name, type);

    // Wrap the variable into a syntax tree node.
    final NodeVariable node = new NodeVariable(variable);
    // Set the back reference to access the reference attributes from the variable.
    node.setUserData(reference);

    return node;
  }
  
  public static NodeVariable getVariable(final VerilogReference reference) {
    return getVariable(reference, new ValueProvider() {
      @Override
      public Data getVariableValue(final Variable variable) {
        return null;
      }
    });
  }

  //================================================================================================
  // Utilities for VerilogDeclaration
  //================================================================================================

  public static String getVariableName(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);
    return declaration.getName();
  }

  public static NodeVariable getVariable(
      final VerilogDeclaration declaration,
      final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(declaration);

    final String name = declaration.getName();
    InvariantChecks.checkNotNull(name,
        String.format("Declaration=%s, parent=%s", declaration, declaration.getParentNode()));

    final DataType type = getDataType(declaration, valueProvider, null);
    InvariantChecks.checkNotNull(type);

    // Create an untyped variable; the data type is refined in the static checker.
    final Variable variable = new Variable(name, type);
    // Wrap the variable into a syntax tree node.
    return new NodeVariable(variable);
  }

  public static NodeVariable getVariable(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);

    return getVariable(declaration, new ValueProvider() {
      @Override
      public Data getVariableValue(final Variable variable) {
        return null;
      }
    });
  }

  private static DataType getIntegerDataType() {
    // Integer are represented as bit vectors of length 32.
    return DataType.bitVector(Integer.SIZE);
  }

  private static int log2(int x) {
    // In the case of floor rounding, n should be 0.
    int n = (x & (x - 1)) != 0 ? 1 : 0;

    if((x & 0xffff0000) != 0) { x >>>= 16; n += 16; }
    if((x & 0xffffff00) != 0) { x >>>=  8; n +=  8; }
    if((x & 0xfffffff0) != 0) { x >>>=  4; n +=  4; }
    if((x & 0xfffffffc) != 0) { x >>>=  2; n +=  2; }

    return n + (x >>> 1);
  }

  public static DataType getDataType(
      final VerilogDeclaration declaration,
      final ValueProvider valueProvider,
      final List<VerilogIndexRange> indexRanges) {
    InvariantChecks.checkNotNull(declaration);
    InvariantChecks.checkNotNull(valueProvider);
    // The parameter indexRanges can be null.

    if (declaration.isGenvar()) {
      return getIntegerDataType();
    }

    if (declaration.isParam()) {
      // FIXME: Parameters may be of different types.
      return getIntegerDataType();
    }

    DataType elementDataType = getDataType(declaration.getElementType(), valueProvider);

    for (final VerilogRange dimension : declaration.getDimensions()) {

      final DataType indexDataType;

      if (indexRanges != null) {
        final VerilogIndexRange indexRange = getIndexRange(dimension, valueProvider);
        indexRanges.add(indexRange);

        final int indexRangeSize = (indexRange.max - indexRange.min) + 1;
        indexDataType = DataType.bitVector(log2(indexRangeSize));
      } else {
        indexDataType = getIntegerDataType();
      }

      elementDataType = DataType.map(indexDataType, elementDataType);
    }

    return elementDataType;
  }

  public static DataType getDataType(
      final VerilogDeclaration declaration,
      final List<VerilogIndexRange> indexRanges) {
    return getDataType(declaration, new ValueProvider() {
      @Override
      public Data getVariableValue(final Variable variable) {
        return null;
      }
    }, indexRanges);
  }

  private static DataType getDataType(
      final VerilogElementType elementType,
      final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(elementType);
    InvariantChecks.checkNotNull(valueProvider);

    switch (elementType.getType()) {
      case INTEGER:
      case TIME:
        return getIntegerDataType();
      case REAL:
      case REALTIME:
        return DataType.REAL;
      default:
        final VerilogRange range = elementType.getRange();
        return range != null ? getDataType(range, valueProvider) : DataType.bitVector(1);
    }
  }

  private static DataType getDataType(
      final VerilogRange range,
      final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(range);
    InvariantChecks.checkNotNull(valueProvider);

    final VerilogExpression lhs = range.getLeftExpression();
    final VerilogExpression rhs = range.getRightExpression();

    switch (range.getType()) {
      case PLUS:
      case MINUS:
        final Integer w = VerilogCalculator.evaluateInteger(rhs.getNode(), valueProvider);
        return w != null ? DataType.bitVector(w) : DataType.UNKNOWN;
      default:
        final Integer l = VerilogCalculator.evaluateInteger(lhs.getNode(), valueProvider);
        final Integer r = VerilogCalculator.evaluateInteger(rhs.getNode(), valueProvider);

        return l != null && r != null ? DataType.bitVector(Math.abs(r - l) + 1) : DataType.UNKNOWN;
    }
  }

  private static VerilogIndexRange getIndexRange(
      final VerilogRange range,
      final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(range);
    InvariantChecks.checkNotNull(valueProvider);

    final int lower = getLowerBit(range, valueProvider);
    final int upper = getUpperBit(range, valueProvider);

    return new VerilogIndexRange(lower, upper);
  }

  public static int getLowerBit(final VerilogRange range, final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(range);

    final VerilogExpression lhs = range.getLeftExpression();
    final VerilogExpression rhs = range.getRightExpression();

    final Integer l = VerilogCalculator.evaluateInteger(lhs.getNode(), valueProvider);
    InvariantChecks.checkNotNull(l,
        String.format("Cannot evaluate node '%s'", lhs.getNode()));

    if (rhs == null) {
      return l;
    }

    final Integer r = VerilogCalculator.evaluateInteger(rhs.getNode(), valueProvider);
    InvariantChecks.checkNotNull(r,
        String.format("Cannot evaluate node '%s'", rhs.getNode()));

    switch (range.getType()) {
      case PLUS:
        return l /* Lower */;
      case MINUS:
        return l /* Upper */ - r /* Width */ + 1;
      default:
        return Math.min(l, r);
    }
  }

  public static int getUpperBit(final VerilogRange range, final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(range);

    final VerilogExpression lhs = range.getLeftExpression();
    final VerilogExpression rhs = range.getRightExpression();

    final Integer l = VerilogCalculator.evaluateInteger(lhs.getNode());
    InvariantChecks.checkNotNull(l,
        String.format("Cannot evaluate node '%s'", lhs.getNode()));

    if (rhs == null) {
      return l;
    }

    final Integer r = VerilogCalculator.evaluateInteger(rhs.getNode());
    InvariantChecks.checkNotNull(r,
        String.format("Cannot evaluate node '%s'", rhs.getNode()));

    switch (range.getType()) {
      case PLUS:
        return l /* Lower */ + r /* Width */ - 1;
      case MINUS:
        return r /* Upper */;
      default:
        return Math.max(l, r);
    }
  }

  public static int getLowerBit(final VerilogRange range) {
    return getLowerBit(range, null);
  }

  public static int getUpperBit(final VerilogRange range) {
    return getUpperBit(range, null);
  }

  /**
   * Returns the set of variables used in the LHS of the assignment.
   * 
   * @param assignment the assignment.
   * @return the set of defined variables.
   */
  public static Set<NodeVariable> getDefinedVariables(final VerilogAssignment assignment) {
    InvariantChecks.checkNotNull(assignment);

    final List<VerilogReference> references = assignment.getReferences();
    final Set<NodeVariable> variables = new HashSet<>();

    for (final VerilogReference reference : references) {
      variables.add(getVariable(reference));
    }

    return variables;
  }

  /**
   * Returns the set of variables used in the RHS of the assignment.
   * 
   * @param assignment the assignment.
   * @return the set of used variables.
   */
  public static Set<NodeVariable> getUsedVariables(final VerilogAssignment assignment) {
    InvariantChecks.checkNotNull(assignment);

    final Set<NodeVariable> variables = new HashSet<>();
    final VerilogMinTypMax expression = assignment.getRhsExpression();

    if (expression != null) {
      final VerilogExpression min = expression.getMinExpression();
      final VerilogExpression typ = expression.getTypExpression();
      final VerilogExpression max = expression.getMaxExpression();

      if (min != null) {
        variables.addAll(ExprUtils.getVariables(min.getNode()));
      }
      if (typ != null) {
        variables.addAll(ExprUtils.getVariables(typ.getNode()));
      }
      if (max != null) {
        variables.addAll(ExprUtils.getVariables(max.getNode()));
      }
    }

    for (final VerilogReference reference : assignment.getReferences()) {
      for (final VerilogRange range : reference.getIndicesAndBitsSelection()) {
        final VerilogExpression lhs = range.getLeftExpression();
        final VerilogExpression rhs = range.getRightExpression();

        if (lhs != null) {
          variables.addAll(ExprUtils.getVariables(lhs.getNode()));
        }
        if (rhs != null) {
          variables.addAll(ExprUtils.getVariables(rhs.getNode()));
        }
      }
    }

    return variables;
  }
}
