/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogNode;

import java.util.Collections;
import java.util.Map;

/**
 * {@link VerilogTransformerBvextract} applies offset to bit-vector extraction operations.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTransformerBvextract extends VerilogTransformer {

  /**
   * Applies the offset to the bit-vector extract operations.
   * 
   * @param node the AST node to be transformed.
   * @param variables the variables' descriptors.
   */
  public static void transform(
      final VerilogNode node,
      final Map<String, VerilogDescriptor> variables) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(variables);

    final VerilogTransformer transformer = new VerilogTransformerBvextract(variables);
    transformer.run(node);
  }

  private final Map<String, VerilogDescriptor> variables;

  public VerilogTransformerBvextract(final Map<String, VerilogDescriptor> variables) {
    InvariantChecks.checkNotNull(variables);
    this.variables = Collections.unmodifiableMap(variables);
  }

  @Override
  public Node transform(final Node node) {
    InvariantChecks.checkNotNull(node);

    return Transformer.transform(node, StandardOperation.BVEXTRACT, new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVEXTRACT);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;

        final Node operand0 = operation.getOperand(0);
        final Node operand1 = operation.getOperand(1);
        final Node operand2 = operation.getOperand(2);

        if (operand2.getKind() != Node.Kind.VARIABLE) {
          return node;
        }

        final NodeVariable variable = (NodeVariable) operand2;

        final VerilogDescriptor descriptor = variables.get(variable.getName());
        InvariantChecks.checkNotNull(descriptor,
            String.format("Descriptor for '%s' has not been found: %s", variable, variables));

        if (descriptor.getOffset() == 0) {
          return node;
        }

        final Node offset = NodeValue.newBitVector(descriptor.getOffset(), Integer.SIZE);
        final Node upper = VerilogCalculator.reduce(Nodes.bvsub(operand0, offset));
        final Node lower = VerilogCalculator.reduce(Nodes.bvsub(operand1, offset));

        return Nodes.bvextract(upper, lower, operand2);
      }
    });
  }
}
